# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from boto3.docs.subresource import SubResourceDocumenter
from tests.unit.docs import BaseDocsTest


class TestSubResourceDocumenter(BaseDocsTest):
    def test_document_sub_resources(self):
        sub_resource_documentor = SubResourceDocumenter(self.resource)
        sub_resource_documentor.document_sub_resources(self.doc_structure)
        self.assert_contains_lines_in_order(
            [
                '.. py:method:: Sample(name)',
                '  Creates a Sample resource.::',
                "    sample = myservice.Sample('name')",
                '  :type name: string',
                "  :param name: The Sample's name identifier.",
                '  :rtype: :py:class:`MyService.Sample`',
                '  :returns: A Sample resource',
            ]
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from boto3.docs.service import ServiceDocumenter
from boto3.session import Session
from tests.functional.docs import BaseDocsFunctionalTests


class TestS3Customizations(BaseDocsFunctionalTests):
    def setUp(self):
        self.documenter = ServiceDocumenter(
            's3', session=Session(region_name='us-east-1')
        )
        self.generated_contents = self.documenter.document_service()
        self.generated_contents = self.generated_contents.decode('utf-8')

    def test_file_transfer_methods_are_documented(self):
        self.assert_contains_lines_in_order(
            [
                '.. py:class:: S3.Client',
                '  *   :py:meth:`~S3.Client.download_file`',
                '  *   :py:meth:`~S3.Client.upload_file`',
                '  .. py:method:: download_file(',
                '  .. py:method:: upload_file(',
            ],
            self.generated_contents,
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from boto3.resources.collection import ResourceCollection
from boto3.session import Session
from tests import unittest


class TestCollection(unittest.TestCase):
    def setUp(self):
        self.session = Session(
            aws_access_key_id='dummy',
            aws_secret_access_key='dummy',
            region_name='us-east-1',
        )
        # Pick an arbitrary resource.
        self.ec2_resource = self.session.resource('ec2')

    def test_can_use_collection_methods(self):
        assert isinstance(
            self.ec2_resource.instances.all(), ResourceCollection
        )

    def test_can_chain_methods(self):
        assert isinstance(
            self.ec2_resource.instances.all().page_size(5), ResourceCollection
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from tests import unittest


class BaseDocsFunctionalTests(unittest.TestCase):
    def assert_contains_lines_in_order(self, lines, contents):
        for line in lines:
            assert line in contents
            beginning = contents.find(line)
            contents = contents[(beginning + len(line)) :]

    def get_class_document_block(self, class_name, contents):
        start_class_document = '.. py:class:: %s' % class_name
        start_index = contents.find(start_class_document)
        assert start_index != -1, 'Class is not found in contents'
        contents = contents[start_index:]
        end_index = contents.find('  .. py:class::', len(start_class_document))
        return contents[:end_index]

    def get_method_document_block(self, method_name, contents):
        start_method_document = '  .. py:method:: %s(' % method_name
        start_index = contents.find(start_method_document)
        assert start_index != -1, 'Method is not found in contents'
        contents = contents[start_index:]
        end_index = contents.find(
            '  .. py:method::', len(start_method_document)
        )
        return contents[:end_index]

    def get_request_syntax_document_block(self, contents):
        start_marker = '**Request Syntax**'
        start_index = contents.find(start_marker)
        assert start_index != -1, 'There is no request syntax section'
        contents = contents[start_index:]
        end_index = contents.find(':type', len(start_marker))
        return contents[:end_index]

    def get_response_syntax_document_block(self, contents):
        start_marker = '**Response Syntax**'
        start_index = contents.find(start_marker)
        assert start_index != -1, 'There is no response syntax section'
        contents = contents[start_index:]
        end_index = contents.find('**Response Structure**', len(start_marker))
        return contents[:end_index]

    def get_request_parameter_document_block(self, param_name, contents):
        start_param_document = ':type %s:' % param_name
        start_index = contents.find(start_param_document)
        assert start_index != -1, 'Param is not found in contents'
        contents = contents[start_index:]
        end_index = contents.find(':type', len(start_param_document))
        return contents[:end_index]

    def get_response_parameter_document_block(self, param_name, contents):
        start_param_document = '**Response Structure**'
        start_index = contents.find(start_param_document)
        assert start_index != -1, 'There is no response structure'

        start_param_document = '- **%s**' % param_name
        start_index = contents.find(start_param_document)
        assert start_index != -1, 'Param is not found in contents'
        contents = contents[start_index:]
        end_index = contents.find('- **', len(start_param_document))
        return contents[:end_index]



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License'). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the 'license' file accompanying this file. This file is
# distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import io

import pytest
from botocore.exceptions import ClientError

from boto3.s3 import inject
from tests import mock, unittest


class TestInjectTransferMethods(unittest.TestCase):
    def test_inject_upload_download_file_to_client(self):
        class_attributes = {}
        inject.inject_s3_transfer_methods(class_attributes=class_attributes)
        assert 'upload_file' in class_attributes
        assert 'download_file' in class_attributes

    def test_upload_file_proxies_to_transfer_object(self):
        with mock.patch('boto3.s3.inject.S3Transfer') as transfer:
            inject.upload_file(
                mock.sentinel.CLIENT,
                Filename='filename',
                Bucket='bucket',
                Key='key',
            )
            transfer_in_context_manager = (
                transfer.return_value.__enter__.return_value
            )
            transfer_in_context_manager.upload_file.assert_called_with(
                filename='filename',
                bucket='bucket',
                key='key',
                extra_args=None,
                callback=None,
            )

    def test_download_file_proxies_to_transfer_object(self):
        with mock.patch('boto3.s3.inject.S3Transfer') as transfer:
            inject.download_file(
                mock.sentinel.CLIENT,
                Bucket='bucket',
                Key='key',
                Filename='filename',
            )
            transfer_in_context_manager = (
                transfer.return_value.__enter__.return_value
            )
            transfer_in_context_manager.download_file.assert_called_with(
                bucket='bucket',
                key='key',
                filename='filename',
                extra_args=None,
                callback=None,
            )


class TestBucketLoad(unittest.TestCase):
    def setUp(self):
        self.client = mock.Mock()
        self.resource = mock.Mock()
        self.resource.meta.client = self.client

    def test_bucket_load_finds_bucket(self):
        self.resource.name = 'MyBucket'
        self.client.list_buckets.return_value = {
            'Buckets': [
                {'Name': 'NotMyBucket', 'CreationDate': 1},
                {'Name': self.resource.name, 'CreationDate': 2},
            ],
        }

        inject.bucket_load(self.resource)
        assert self.resource.meta.data == {
            'Name': self.resource.name,
            'CreationDate': 2,
        }

    def test_bucket_load_doesnt_find_bucket(self):
        self.resource.name = 'MyBucket'
        self.client.list_buckets.return_value = {
            'Buckets': [
                {'Name': 'NotMyBucket', 'CreationDate': 1},
                {'Name': 'NotMine2', 'CreationDate': 2},
            ],
        }
        inject.bucket_load(self.resource)
        assert self.resource.meta.data == {}

    def test_bucket_load_encounters_access_exception(self):
        self.client.list_buckets.side_effect = ClientError(
            {'Error': {'Code': 'AccessDenied', 'Message': 'Access Denied'}},
            'ListBuckets',
        )
        inject.bucket_load(self.resource)
        assert self.resource.meta.data == {}

    def test_bucket_load_encounters_other_exception(self):
        self.client.list_buckets.side_effect = ClientError(
            {
                'Error': {
                    'Code': 'ExpiredToken',
                    'Message': 'The provided token has expired.',
                }
            },
            'ListBuckets',
        )
        with pytest.raises(ClientError):
            inject.bucket_load(self.resource)


class TestBucketTransferMethods(unittest.TestCase):
    def setUp(self):
        self.bucket = mock.Mock(name='my_bucket')
        self.copy_source = {'Bucket': 'foo', 'Key': 'bar'}

    def test_upload_file_proxies_to_meta_client(self):
        inject.bucket_upload_file(self.bucket, Filename='foo', Key='key')
        self.bucket.meta.client.upload_file.assert_called_with(
            Filename='foo',
            Bucket=self.bucket.name,
            Key='key',
            ExtraArgs=None,
            Callback=None,
            Config=None,
        )

    def test_download_file_proxies_to_meta_client(self):
        inject.bucket_download_file(self.bucket, Key='key', Filename='foo')
        self.bucket.meta.client.download_file.assert_called_with(
            Bucket=self.bucket.name,
            Key='key',
            Filename='foo',
            ExtraArgs=None,
            Callback=None,
            Config=None,
        )

    def test_copy(self):
        inject.bucket_copy(self.bucket, self.copy_source, Key='key')
        self.bucket.meta.client.copy.assert_called_with(
            CopySource=self.copy_source,
            Bucket=self.bucket.name,
            Key='key',
            ExtraArgs=None,
            Callback=None,
            SourceClient=None,
            Config=None,
        )

    def test_upload_fileobj(self):
        fileobj = io.BytesIO(b'foo')
        inject.bucket_upload_fileobj(self.bucket, Key='key', Fileobj=fileobj)
        self.bucket.meta.client.upload_fileobj.assert_called_with(
            Bucket=self.bucket.name,
            Fileobj=fileobj,
            Key='key',
            ExtraArgs=None,
            Callback=None,
            Config=None,
        )

    def test_download_fileobj(self):
        obj = io.BytesIO()
        inject.bucket_download_fileobj(self.bucket, Key='key', Fileobj=obj)
        self.bucket.meta.client.download_fileobj.assert_called_with(
            Bucket=self.bucket.name,
            Key='key',
            Fileobj=obj,
            ExtraArgs=None,
            Callback=None,
            Config=None,
        )


class TestObjectTransferMethods(unittest.TestCase):
    def setUp(self):
        self.obj = mock.Mock(bucket_name='my_bucket', key='my_key')
        self.copy_source = {'Bucket': 'foo', 'Key': 'bar'}

    def test_upload_file_proxies_to_meta_client(self):
        inject.object_upload_file(self.obj, Filename='foo')
        self.obj.meta.client.upload_file.assert_called_with(
            Filename='foo',
            Bucket=self.obj.bucket_name,
            Key=self.obj.key,
            ExtraArgs=None,
            Callback=None,
            Config=None,
        )

    def test_download_file_proxies_to_meta_client(self):
        inject.object_download_file(self.obj, Filename='foo')
        self.obj.meta.client.download_file.assert_called_with(
            Bucket=self.obj.bucket_name,
            Key=self.obj.key,
            Filename='foo',
            ExtraArgs=None,
            Callback=None,
            Config=None,
        )

    def test_copy(self):
        inject.object_copy(self.obj, self.copy_source)
        self.obj.meta.client.copy.assert_called_with(
            CopySource=self.copy_source,
            Bucket=self.obj.bucket_name,
            Key=self.obj.key,
            ExtraArgs=None,
            Callback=None,
            SourceClient=None,
            Config=None,
        )

    def test_upload_fileobj(self):
        fileobj = io.BytesIO(b'foo')
        inject.object_upload_fileobj(self.obj, Fileobj=fileobj)
        self.obj.meta.client.upload_fileobj.assert_called_with(
            Bucket=self.obj.bucket_name,
            Fileobj=fileobj,
            Key=self.obj.key,
            ExtraArgs=None,
            Callback=None,
            Config=None,
        )

    def test_download_fileobj(self):
        fileobj = io.BytesIO()
        inject.object_download_fileobj(self.obj, Fileobj=fileobj)
        self.obj.meta.client.download_fileobj.assert_called_with(
            Bucket=self.obj.bucket_name,
            Key=self.obj.key,
            Fileobj=fileobj,
            ExtraArgs=None,
            Callback=None,
            Config=None,
        )


class TestObejctSummaryLoad(unittest.TestCase):
    def setUp(self):
        self.client = mock.Mock()
        self.resource = mock.Mock()
        self.resource.meta.client = self.client
        self.head_object_response = {'ContentLength': 5, 'ETag': 'my-etag'}
        self.client.head_object.return_value = self.head_object_response

    def test_object_summary_load(self):
        inject.object_summary_load(self.resource)
        assert self.resource.meta.data == {'Size': 5, 'ETag': 'my-etag'}

    def test_can_handle_missing_content_length(self):
        self.head_object_response.pop('ContentLength')
        inject.object_summary_load(self.resource)
        assert self.resource.meta.data == {'ETag': 'my-etag'}



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import botocore.session
import pytest
from botocore import xform_name
from botocore.exceptions import DataNotFoundError

import boto3
from boto3.docs.service import ServiceDocumenter


@pytest.fixture
def botocore_session():
    return botocore.session.get_session()


@pytest.fixture
def boto3_session():
    return boto3.Session(region_name='us-east-1')


def all_services():
    session = boto3.Session(region_name='us-east-1')
    yield from session.get_available_services()


@pytest.fixture
def available_resources():
    session = boto3.Session(region_name='us-east-1')
    return session.get_available_resources()


@pytest.mark.parametrize('service_name', all_services())
def test_documentation(
    boto3_session, botocore_session, available_resources, service_name
):
    generated_docs = ServiceDocumenter(
        service_name, session=boto3_session
    ).document_service()
    generated_docs = generated_docs.decode('utf-8')
    client = boto3.client(service_name, 'us-east-1')

    # Check that all of the services have the appropriate title
    _assert_has_title(generated_docs, client)

    # Check that all services have the client documented.
    _assert_has_client_documentation(generated_docs, service_name, client)

    # If the service has resources, make sure the service resource
    # is at least documented.
    if service_name in available_resources:

        resource = boto3.resource(service_name, 'us-east-1')
        _assert_has_resource_documentation(
            generated_docs, service_name, resource
        )

    # If the client can paginate, make sure the paginators are documented.
    try:
        paginator_model = botocore_session.get_paginator_model(service_name)
        _assert_has_paginator_documentation(
            generated_docs,
            service_name,
            client,
            sorted(paginator_model._paginator_config),
        )
    except DataNotFoundError:
        pass

    # If the client has waiters, make sure the waiters are documented.
    if client.waiter_names:
        waiter_model = botocore_session.get_waiter_model(service_name)
        _assert_has_waiter_documentation(
            generated_docs, service_name, client, waiter_model
        )


def _assert_contains_lines_in_order(lines, contents):
    for line in lines:
        assert line in contents
        beginning = contents.find(line)
        contents = contents[(beginning + len(line)) :]


def _assert_has_title(generated_docs, client):
    title = client.__class__.__name__
    ref_lines = ['*' * len(title), title, '*' * len(title)]
    _assert_contains_lines_in_order(ref_lines, generated_docs)


def _assert_has_client_documentation(generated_docs, service_name, client):
    class_name = client.__class__.__name__
    ref_lines = [
        '======',
        'Client',
        '======',
        '.. py:class:: %s.Client' % class_name,
        '  A low-level client representing',
        '    import boto3',
        '    client = boto3.client(\'%s\')' % service_name,
        '  These are the available methods:',
        '  *   :py:meth:`~%s.Client.get_paginator`' % class_name,
        '  *   :py:meth:`~%s.Client.get_waiter`' % class_name,
        '  .. py:method:: get_paginator(operation_name)',
        '  .. py:method:: get_waiter(waiter_name)',
    ]
    _assert_contains_lines_in_order(ref_lines, generated_docs)


def _assert_has_paginator_documentation(
    generated_docs, service_name, client, paginator_names
):
    ref_lines = [
        '==========',
        'Paginators',
        '==========',
        'The available paginators are:',
    ]
    for paginator_name in paginator_names:
        ref_lines.append(
            '* :py:class:`{}.Paginator.{}`'.format(
                client.__class__.__name__, paginator_name
            )
        )

    for paginator_name in paginator_names:
        ref_lines.append(
            '.. py:class:: {}.Paginator.{}'.format(
                client.__class__.__name__, paginator_name
            )
        )
        ref_lines.append('  .. py:method:: paginate(')

    _assert_contains_lines_in_order(ref_lines, generated_docs)


def _assert_has_waiter_documentation(
    generated_docs, service_name, client, waiter_model
):
    ref_lines = ['=======', 'Waiters', '=======', 'The available waiters are:']
    for waiter_name in waiter_model.waiter_names:
        ref_lines.append(
            '* :py:class:`{}.Waiter.{}`'.format(
                client.__class__.__name__, waiter_name
            )
        )

    for waiter_name in waiter_model.waiter_names:
        ref_lines.append(
            '.. py:class:: {}.Waiter.{}'.format(
                client.__class__.__name__, waiter_name
            )
        )
        ref_lines.append(
            '    waiter = client.get_waiter(\'%s\')' % xform_name(waiter_name)
        )
        ref_lines.append('  .. py:method:: wait(')

    _assert_contains_lines_in_order(ref_lines, generated_docs)


def _assert_has_resource_documentation(generated_docs, service_name, resource):
    ref_lines = [
        '================',
        'Service Resource',
        '================',
        '.. py:class:: %s.ServiceResource'
        % (resource.meta.client.__class__.__name__),
        '  A resource representing',
        '    import boto3',
        f'    {service_name} = boto3.resource(\'{service_name}\')',
    ]
    _assert_contains_lines_in_order(ref_lines, generated_docs)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest
from botocore import loaders
from botocore.client import Config
from botocore.exceptions import UnknownServiceError

from boto3 import __version__
from boto3.exceptions import ResourceNotExistsError
from boto3.session import Session
from tests import BaseTestCase, mock


class TestSession(BaseTestCase):
    def test_repr(self):
        bc_session = self.bc_session_cls.return_value
        bc_session.get_credentials.return_value.access_key = 'abc123'
        bc_session.get_config_variable.return_value = 'us-west-2'

        session = Session('abc123', region_name='us-west-2')

        assert repr(session) == 'Session(region_name=\'us-west-2\')'

    def test_repr_on_subclasses(self):
        bc_session = self.bc_session_cls.return_value
        bc_session.get_credentials.return_value.access_key = 'abc123'
        bc_session.get_config_variable.return_value = 'us-west-2'

        class MySession(Session):
            pass

        session = MySession('abc123', region_name='us-west-2')

        assert repr(session) == 'MySession(region_name=\'us-west-2\')'

    def test_can_access_region_name(self):
        bc_session = self.bc_session_cls.return_value
        bc_session.get_config_variable.return_value = 'us-west-2'
        session = Session('abc123', region_name='us-west-2')
        bc_session.set_config_variable.assert_called_with(
            'region', 'us-west-2'
        )

        assert session.region_name == 'us-west-2'

    def test_arguments_not_required(self):
        Session()

        assert self.bc_session_cls.called

    def test_credentials_can_be_set(self):
        bc_session = self.bc_session_cls.return_value

        # Set values in constructor
        Session(
            aws_access_key_id='key',
            aws_secret_access_key='secret',
            aws_session_token='token',
        )

        assert self.bc_session_cls.called
        assert bc_session.set_credentials.called
        bc_session.set_credentials.assert_called_with('key', 'secret', 'token')

    def test_can_get_credentials(self):
        access_key = 'foo'
        secret_key = 'bar'
        token = 'baz'

        creds = mock.Mock()
        creds.access_key = access_key
        creds.secret_key = secret_key
        creds.token = token

        bc_session = self.bc_session_cls.return_value
        bc_session.get_credentials.return_value = creds

        session = Session(
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            aws_session_token=token,
        )

        credentials = session.get_credentials()
        assert credentials.access_key == access_key
        assert credentials.secret_key == secret_key
        assert credentials.token == token

    def test_profile_can_be_set(self):
        bc_session = self.bc_session_cls.return_value

        session = Session(profile_name='foo')

        bc_session.set_config_variable.assert_called_with('profile', 'foo')
        bc_session.profile = 'foo'

        # We should also be able to read the value
        assert session.profile_name == 'foo'

    def test_profile_default(self):
        self.bc_session_cls.return_value.profile = None

        session = Session()

        assert session.profile_name == 'default'

    def test_available_profiles(self):
        bc_session = mock.Mock()
        bc_session.available_profiles.return_value = ['foo', 'bar']
        session = Session(botocore_session=bc_session)
        profiles = session.available_profiles
        assert len(profiles.return_value) == 2

    def test_custom_session(self):
        bc_session = self.bc_session_cls()
        self.bc_session_cls.reset_mock()

        Session(botocore_session=bc_session)

        # No new session was created
        assert not self.bc_session_cls.called

    def test_user_agent(self):
        # Here we get the underlying Botocore session, create a Boto3
        # session, and ensure that the user-agent is modified as expected
        bc_session = self.bc_session_cls.return_value
        bc_session.user_agent_name = 'Botocore'
        bc_session.user_agent_version = '0.68.0'
        bc_session.user_agent_extra = ''

        Session(botocore_session=bc_session)

        assert bc_session.user_agent_name == 'Boto3'
        assert bc_session.user_agent_version == __version__
        assert bc_session.user_agent_extra == 'Botocore/0.68.0'

    def test_user_agent_extra(self):
        # This test is the same as above, but includes custom extra content
        # which must still be in the final modified user-agent.
        bc_session = self.bc_session_cls.return_value
        bc_session.user_agent_name = 'Botocore'
        bc_session.user_agent_version = '0.68.0'
        bc_session.user_agent_extra = 'foo'

        Session(botocore_session=bc_session)

        assert bc_session.user_agent_extra == 'foo Botocore/0.68.0'

    def test_custom_user_agent(self):
        # This test ensures that a customized user-agent is left untouched.
        bc_session = self.bc_session_cls.return_value
        bc_session.user_agent_name = 'Custom'
        bc_session.user_agent_version = '1.0'
        bc_session.user_agent_extra = ''

        Session(botocore_session=bc_session)

        assert bc_session.user_agent_name == 'Custom'
        assert bc_session.user_agent_version == '1.0'
        assert bc_session.user_agent_extra == ''

    def test_get_available_services(self):
        bc_session = self.bc_session_cls.return_value

        session = Session()
        session.get_available_services()

        assert bc_session.get_available_services.called

    def test_get_available_resources(self):
        mock_bc_session = mock.Mock()
        loader = mock.Mock(spec=loaders.Loader)
        loader.list_available_services.return_value = ['foo', 'bar']
        mock_bc_session.get_component.return_value = loader
        session = Session(botocore_session=mock_bc_session)

        names = session.get_available_resources()
        assert names == ['foo', 'bar']

    def test_get_available_partitions(self):
        bc_session = mock.Mock()
        bc_session.get_available_partitions.return_value = ['foo']
        session = Session(botocore_session=bc_session)

        partitions = session.get_available_partitions()
        assert partitions == ['foo']

    def test_get_available_regions(self):
        bc_session = mock.Mock()
        bc_session.get_available_regions.return_value = ['foo']
        session = Session(botocore_session=bc_session)

        partitions = session.get_available_regions('myservice')
        bc_session.get_available_regions.assert_called_with(
            service_name='myservice',
            partition_name='aws',
            allow_non_regional=False,
        )
        assert partitions == ['foo']

    def test_get_partition_for_region(self):
        bc_session = mock.Mock()
        bc_session.get_partition_for_region.return_value = 'baz'
        session = Session(botocore_session=bc_session)

        partition = session.get_partition_for_region('foo-bar-1')
        bc_session.get_partition_for_region.assert_called_with('foo-bar-1')
        assert partition == 'baz'

    def test_create_client(self):
        session = Session(region_name='us-east-1')
        client = session.client('sqs', region_name='us-west-2')

        assert client, 'No low-level client was returned'

    def test_create_client_with_args(self):
        bc_session = self.bc_session_cls.return_value

        session = Session(region_name='us-east-1')
        session.client('sqs', region_name='us-west-2')

        bc_session.create_client.assert_called_with(
            'sqs',
            aws_secret_access_key=None,
            aws_access_key_id=None,
            endpoint_url=None,
            use_ssl=True,
            aws_session_token=None,
            verify=None,
            region_name='us-west-2',
            api_version=None,
            config=None,
        )

    def test_create_resource_with_args(self):
        mock_bc_session = mock.Mock()
        loader = mock.Mock(spec=loaders.Loader)
        loader.determine_latest_version.return_value = '2014-11-02'
        loader.load_service_model.return_value = {
            'resources': [],
            'service': [],
        }
        mock_bc_session.get_component.return_value = loader
        session = Session(botocore_session=mock_bc_session)
        session.resource_factory.load_from_definition = mock.Mock()
        session.client = mock.Mock()

        session.resource('sqs', verify=False)

        session.client.assert_called_with(
            'sqs',
            aws_secret_access_key=None,
            aws_access_key_id=None,
            endpoint_url=None,
            use_ssl=True,
            aws_session_token=None,
            verify=False,
            region_name=None,
            api_version='2014-11-02',
            config=mock.ANY,
        )
        client_config = session.client.call_args[1]['config']
        assert client_config.user_agent_extra == 'Resource'
        assert client_config.signature_version is None

    def test_create_resource_with_config(self):
        mock_bc_session = mock.Mock()
        loader = mock.Mock(spec=loaders.Loader)
        loader.determine_latest_version.return_value = '2014-11-02'
        loader.load_service_model.return_value = {
            'resources': [],
            'service': [],
        }
        mock_bc_session.get_component.return_value = loader
        session = Session(botocore_session=mock_bc_session)
        session.resource_factory.load_from_definition = mock.Mock()
        session.client = mock.Mock()
        config = Config(signature_version='v4')

        session.resource('sqs', config=config)

        session.client.assert_called_with(
            'sqs',
            aws_secret_access_key=None,
            aws_access_key_id=None,
            endpoint_url=None,
            use_ssl=True,
            aws_session_token=None,
            verify=None,
            region_name=None,
            api_version='2014-11-02',
            config=mock.ANY,
        )
        client_config = session.client.call_args[1]['config']
        assert client_config.user_agent_extra == 'Resource'
        assert client_config.signature_version == 'v4'

    def test_create_resource_with_config_override_user_agent_extra(self):
        mock_bc_session = mock.Mock()
        loader = mock.Mock(spec=loaders.Loader)
        loader.determine_latest_version.return_value = '2014-11-02'
        loader.load_service_model.return_value = {
            'resources': [],
            'service': [],
        }
        mock_bc_session.get_component.return_value = loader
        session = Session(botocore_session=mock_bc_session)
        session.resource_factory.load_from_definition = mock.Mock()
        session.client = mock.Mock()
        config = Config(signature_version='v4', user_agent_extra='foo')

        session.resource('sqs', config=config)

        session.client.assert_called_with(
            'sqs',
            aws_secret_access_key=None,
            aws_access_key_id=None,
            endpoint_url=None,
            use_ssl=True,
            aws_session_token=None,
            verify=None,
            region_name=None,
            api_version='2014-11-02',
            config=mock.ANY,
        )
        client_config = session.client.call_args[1]['config']
        assert client_config.user_agent_extra == 'foo'
        assert client_config.signature_version == 'v4'

    def test_create_resource_latest_version(self):
        mock_bc_session = mock.Mock()
        loader = mock.Mock(spec=loaders.Loader)
        loader.determine_latest_version.return_value = '2014-11-02'
        loader.load_service_model.return_value = {
            'resources': [],
            'service': [],
        }
        mock_bc_session.get_component.return_value = loader
        session = Session(botocore_session=mock_bc_session)
        session.resource_factory.load_from_definition = mock.Mock()

        session.resource('sqs')

        loader.load_service_model.assert_called_with(
            'sqs', 'resources-1', None
        )

    def test_bad_resource_name(self):
        mock_bc_session = mock.Mock()
        loader = mock.Mock(spec=loaders.Loader)
        loader.load_service_model.side_effect = UnknownServiceError(
            service_name='foo', known_service_names='asdf'
        )
        mock_bc_session.get_component.return_value = loader
        loader.list_available_services.return_value = ['good-resource']
        mock_bc_session.get_available_services.return_value = ['sqs']

        session = Session(botocore_session=mock_bc_session)
        with pytest.raises(ResourceNotExistsError) as e:
            session.resource('sqs')

            err_msg = str(e.exception)
            # 1. should say the resource doesn't exist.
            assert 'resource does not exist' in err_msg
            assert 'sqs' in err_msg
            # 2. Should list available resources you can choose.
            assert 'good-resource' in err_msg
            # 3. Should list client if available.
            assert 'client' in err_msg

    def test_bad_resource_name_with_no_client_has_simple_err_msg(self):
        mock_bc_session = mock.Mock()
        loader = mock.Mock(spec=loaders.Loader)
        loader.load_service_model.side_effect = UnknownServiceError(
            service_name='foo', known_service_names='asdf'
        )
        mock_bc_session.get_component.return_value = loader
        loader.list_available_services.return_value = ['good-resource']
        mock_bc_session.get_available_services.return_value = ['good-client']

        session = Session(botocore_session=mock_bc_session)
        with pytest.raises(ResourceNotExistsError) as e:
            session.resource('bad-client')

            err_msg = str(e.exception)
            # Shouldn't mention anything about clients because
            # 'bad-client' it not a valid boto3.client(...)
            assert 'boto3.client' not in err_msg

    def test_can_reach_events(self):
        mock_bc_session = self.bc_session_cls()
        session = Session(botocore_session=mock_bc_session)
        session.events
        mock_bc_session.get_component.assert_called_with('event_emitter')



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.model import OperationModel, ServiceModel

from boto3.dynamodb.conditions import Attr, Key
from boto3.dynamodb.transform import (
    DynamoDBHighLevelResource,
    ParameterTransformer,
    TransformationInjector,
    copy_dynamodb_params,
    register_high_level_interface,
)
from boto3.resources.base import ResourceMeta, ServiceResource
from tests import mock, unittest


class BaseTransformationTest(unittest.TestCase):
    def setUp(self):
        self.target_shape = 'MyShape'
        self.original_value = 'orginal'
        self.transformed_value = 'transformed'
        self.transformer = ParameterTransformer()
        self.json_model = {}
        self.nested_json_model = {}
        self.setup_models()
        self.build_models()

    def setup_models(self):
        self.json_model = {
            'operations': {
                'SampleOperation': {
                    'name': 'SampleOperation',
                    'input': {'shape': 'SampleOperationInputOutput'},
                    'output': {'shape': 'SampleOperationInputOutput'},
                }
            },
            'shapes': {
                'SampleOperationInputOutput': {
                    'type': 'structure',
                    'members': {},
                },
                'String': {'type': 'string'},
            },
        }

    def build_models(self):
        self.service_model = ServiceModel(self.json_model)
        self.operation_model = OperationModel(
            self.json_model['operations']['SampleOperation'],
            self.service_model,
        )

    def add_input_shape(self, shape):
        self.add_shape(shape)
        params_shape = self.json_model['shapes']['SampleOperationInputOutput']
        shape_name = list(shape.keys())[0]
        params_shape['members'][shape_name] = {'shape': shape_name}

    def add_shape(self, shape):
        shape_name = list(shape.keys())[0]
        self.json_model['shapes'][shape_name] = shape[shape_name]


class TestInputOutputTransformer(BaseTransformationTest):
    def setUp(self):
        super().setUp()
        self.transformation = lambda params: self.transformed_value
        self.add_shape({self.target_shape: {'type': 'string'}})

    def test_transform_structure(self):
        input_params = {
            'Structure': {
                'TransformMe': self.original_value,
                'LeaveAlone': self.original_value,
            }
        }
        input_shape = {
            'Structure': {
                'type': 'structure',
                'members': {
                    'TransformMe': {'shape': self.target_shape},
                    'LeaveAlone': {'shape': 'String'},
                },
            }
        }

        self.add_input_shape(input_shape)
        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        assert input_params == {
            'Structure': {
                'TransformMe': self.transformed_value,
                'LeaveAlone': self.original_value,
            }
        }

    def test_transform_map(self):
        input_params = {
            'TransformMe': {'foo': self.original_value},
            'LeaveAlone': {'foo': self.original_value},
        }

        targeted_input_shape = {
            'TransformMe': {
                'type': 'map',
                'key': {'shape': 'String'},
                'value': {'shape': self.target_shape},
            }
        }

        untargeted_input_shape = {
            'LeaveAlone': {
                'type': 'map',
                'key': {'shape': 'String'},
                'value': {'shape': 'String'},
            }
        }

        self.add_input_shape(targeted_input_shape)
        self.add_input_shape(untargeted_input_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        assert input_params == {
            'TransformMe': {'foo': self.transformed_value},
            'LeaveAlone': {'foo': self.original_value},
        }

    def test_transform_list(self):
        input_params = {
            'TransformMe': [self.original_value, self.original_value],
            'LeaveAlone': [self.original_value, self.original_value],
        }

        targeted_input_shape = {
            'TransformMe': {
                'type': 'list',
                'member': {'shape': self.target_shape},
            }
        }

        untargeted_input_shape = {
            'LeaveAlone': {'type': 'list', 'member': {'shape': 'String'}}
        }

        self.add_input_shape(targeted_input_shape)
        self.add_input_shape(untargeted_input_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        assert input_params == {
            'TransformMe': [self.transformed_value, self.transformed_value],
            'LeaveAlone': [self.original_value, self.original_value],
        }

    def test_transform_nested_structure(self):
        input_params = {
            'WrapperStructure': {
                'Structure': {
                    'TransformMe': self.original_value,
                    'LeaveAlone': self.original_value,
                }
            }
        }

        structure_shape = {
            'Structure': {
                'type': 'structure',
                'members': {
                    'TransformMe': {'shape': self.target_shape},
                    'LeaveAlone': {'shape': 'String'},
                },
            }
        }

        input_shape = {
            'WrapperStructure': {
                'type': 'structure',
                'members': {'Structure': {'shape': 'Structure'}},
            }
        }
        self.add_shape(structure_shape)
        self.add_input_shape(input_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        assert input_params == {
            'WrapperStructure': {
                'Structure': {
                    'TransformMe': self.transformed_value,
                    'LeaveAlone': self.original_value,
                }
            }
        }

    def test_transform_nested_map(self):
        input_params = {
            'TargetedWrapperMap': {'foo': {'bar': self.original_value}},
            'UntargetedWrapperMap': {'foo': {'bar': self.original_value}},
        }

        targeted_map_shape = {
            'TransformMeMap': {
                'type': 'map',
                'key': {'shape': 'String'},
                'value': {'shape': self.target_shape},
            }
        }

        targeted_wrapper_shape = {
            'TargetedWrapperMap': {
                'type': 'map',
                'key': {'shape': 'Name'},
                'value': {'shape': 'TransformMeMap'},
            }
        }

        self.add_shape(targeted_map_shape)
        self.add_input_shape(targeted_wrapper_shape)

        untargeted_map_shape = {
            'LeaveAloneMap': {
                'type': 'map',
                'key': {'shape': 'String'},
                'value': {'shape': 'String'},
            }
        }

        untargeted_wrapper_shape = {
            'UntargetedWrapperMap': {
                'type': 'map',
                'key': {'shape': 'Name'},
                'value': {'shape': 'LeaveAloneMap'},
            }
        }

        self.add_shape(untargeted_map_shape)
        self.add_input_shape(untargeted_wrapper_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        assert input_params == {
            'TargetedWrapperMap': {'foo': {'bar': self.transformed_value}},
            'UntargetedWrapperMap': {'foo': {'bar': self.original_value}},
        }

    def test_transform_nested_list(self):
        input_params = {
            'TargetedWrapperList': [
                [self.original_value, self.original_value]
            ],
            'UntargetedWrapperList': [
                [self.original_value, self.original_value]
            ],
        }

        targeted_list_shape = {
            'TransformMe': {
                'type': 'list',
                'member': {'shape': self.target_shape},
            }
        }

        targeted_wrapper_shape = {
            'TargetedWrapperList': {
                'type': 'list',
                'member': {'shape': 'TransformMe'},
            }
        }

        self.add_shape(targeted_list_shape)
        self.add_input_shape(targeted_wrapper_shape)

        untargeted_list_shape = {
            'LeaveAlone': {'type': 'list', 'member': {'shape': 'String'}}
        }

        untargeted_wrapper_shape = {
            'UntargetedWrapperList': {
                'type': 'list',
                'member': {'shape': 'LeaveAlone'},
            }
        }

        self.add_shape(untargeted_list_shape)
        self.add_input_shape(untargeted_wrapper_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        assert input_params == {
            'TargetedWrapperList': [
                [self.transformed_value, self.transformed_value]
            ],
            'UntargetedWrapperList': [
                [self.original_value, self.original_value]
            ],
        }

    def test_transform_incorrect_type_for_structure(self):
        input_params = {'Structure': 'foo'}

        input_shape = {
            'Structure': {
                'type': 'structure',
                'members': {
                    'TransformMe': {'shape': self.target_shape},
                },
            }
        }

        self.add_input_shape(input_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        assert input_params == {'Structure': 'foo'}

    def test_transform_incorrect_type_for_map(self):
        input_params = {'Map': 'foo'}

        input_shape = {
            'Map': {
                'type': 'map',
                'key': {'shape': 'String'},
                'value': {'shape': self.target_shape},
            }
        }

        self.add_input_shape(input_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        assert input_params == {'Map': 'foo'}

    def test_transform_incorrect_type_for_list(self):
        input_params = {'List': 'foo'}

        input_shape = {
            'List': {'type': 'list', 'member': {'shape': self.target_shape}}
        }

        self.add_input_shape(input_shape)

        self.transformer.transform(
            params=input_params,
            model=self.operation_model.input_shape,
            transformation=self.transformation,
            target_shape=self.target_shape,
        )
        assert input_params == {'List': 'foo'}


class BaseTransformAttributeValueTest(BaseTransformationTest):
    def setUp(self):
        self.target_shape = 'AttributeValue'
        self.setup_models()
        self.build_models()
        self.python_value = 'mystring'
        self.dynamodb_value = {'S': self.python_value}
        self.injector = TransformationInjector()
        self.add_shape({self.target_shape: {'type': 'string'}})


class TestTransformAttributeValueInput(BaseTransformAttributeValueTest):
    def test_handler(self):
        input_params = {
            'Structure': {
                'TransformMe': self.python_value,
                'LeaveAlone': 'unchanged',
            }
        }
        input_shape = {
            'Structure': {
                'type': 'structure',
                'members': {
                    'TransformMe': {'shape': self.target_shape},
                    'LeaveAlone': {'shape': 'String'},
                },
            }
        }

        self.add_input_shape(input_shape)

        self.injector.inject_attribute_value_input(
            params=input_params, model=self.operation_model
        )
        assert input_params == {
            'Structure': {
                'TransformMe': self.dynamodb_value,
                'LeaveAlone': 'unchanged',
            }
        }


class TestTransformAttributeValueOutput(BaseTransformAttributeValueTest):
    def test_handler(self):
        parsed = {
            'Structure': {
                'TransformMe': self.dynamodb_value,
                'LeaveAlone': 'unchanged',
            }
        }
        input_shape = {
            'Structure': {
                'type': 'structure',
                'members': {
                    'TransformMe': {'shape': self.target_shape},
                    'LeaveAlone': {'shape': 'String'},
                },
            }
        }

        self.add_input_shape(input_shape)
        self.injector.inject_attribute_value_output(
            parsed=parsed, model=self.operation_model
        )
        assert parsed == {
            'Structure': {
                'TransformMe': self.python_value,
                'LeaveAlone': 'unchanged',
            }
        }

    def test_no_output(self):
        service_model = ServiceModel(
            {
                'operations': {
                    'SampleOperation': {
                        'name': 'SampleOperation',
                        'input': {'shape': 'SampleOperationInputOutput'},
                    }
                },
                'shapes': {
                    'SampleOperationInput': {
                        'type': 'structure',
                        'members': {},
                    },
                    'String': {'type': 'string'},
                },
            }
        )
        operation_model = service_model.operation_model('SampleOperation')

        parsed = {}
        self.injector.inject_attribute_value_output(
            parsed=parsed, model=operation_model
        )
        assert parsed == {}


class TestTransformConditionExpression(BaseTransformationTest):
    def setUp(self):
        super().setUp()
        self.add_shape({'ConditionExpression': {'type': 'string'}})
        self.add_shape({'KeyExpression': {'type': 'string'}})

        shapes = self.json_model['shapes']
        input_members = shapes['SampleOperationInputOutput']['members']
        input_members['KeyCondition'] = {'shape': 'KeyExpression'}
        input_members['AttrCondition'] = {'shape': 'ConditionExpression'}
        self.injector = TransformationInjector()
        self.build_models()

    def test_non_condition_input(self):
        params = {'KeyCondition': 'foo', 'AttrCondition': 'bar'}
        self.injector.inject_condition_expressions(
            params, self.operation_model
        )
        assert params == {'KeyCondition': 'foo', 'AttrCondition': 'bar'}

    def test_single_attr_condition_expression(self):
        params = {'AttrCondition': Attr('foo').eq('bar')}
        self.injector.inject_condition_expressions(
            params, self.operation_model
        )
        assert params == {
            'AttrCondition': '#n0 = :v0',
            'ExpressionAttributeNames': {'#n0': 'foo'},
            'ExpressionAttributeValues': {':v0': 'bar'},
        }

    def test_single_key_conditon_expression(self):
        params = {'KeyCondition': Key('foo').eq('bar')}
        self.injector.inject_condition_expressions(
            params, self.operation_model
        )
        assert params == {
            'KeyCondition': '#n0 = :v0',
            'ExpressionAttributeNames': {'#n0': 'foo'},
            'ExpressionAttributeValues': {':v0': 'bar'},
        }

    def test_key_and_attr_conditon_expression(self):
        params = {
            'KeyCondition': Key('foo').eq('bar'),
            'AttrCondition': Attr('biz').eq('baz'),
        }
        self.injector.inject_condition_expressions(
            params, self.operation_model
        )
        assert params == {
            'KeyCondition': '#n1 = :v1',
            'AttrCondition': '#n0 = :v0',
            'ExpressionAttributeNames': {'#n0': 'biz', '#n1': 'foo'},
            'ExpressionAttributeValues': {':v0': 'baz', ':v1': 'bar'},
        }

    def test_key_and_attr_conditon_expression_with_placeholders(self):
        params = {
            'KeyCondition': Key('foo').eq('bar'),
            'AttrCondition': Attr('biz').eq('baz'),
            'ExpressionAttributeNames': {'#a': 'b'},
            'ExpressionAttributeValues': {':c': 'd'},
        }
        self.injector.inject_condition_expressions(
            params, self.operation_model
        )
        assert params == {
            'KeyCondition': '#n1 = :v1',
            'AttrCondition': '#n0 = :v0',
            'ExpressionAttributeNames': {
                '#n0': 'biz',
                '#n1': 'foo',
                '#a': 'b',
            },
            'ExpressionAttributeValues': {
                ':v0': 'baz',
                ':v1': 'bar',
                ':c': 'd',
            },
        }


class TestCopyDynamoDBParams(unittest.TestCase):
    def test_copy_dynamodb_params(self):
        params = {'foo': 'bar'}
        new_params = copy_dynamodb_params(params)
        assert params == new_params
        assert new_params is not params


class TestDynamoDBHighLevelResource(unittest.TestCase):
    def setUp(self):
        self.events = mock.Mock()
        self.client = mock.Mock()
        self.client.meta.events = self.events
        self.meta = ResourceMeta('dynamodb')

    def test_instantiation(self):
        # Instantiate the class.
        dynamodb_class = type(
            'dynamodb',
            (DynamoDBHighLevelResource, ServiceResource),
            {'meta': self.meta},
        )
        with mock.patch(
            'boto3.dynamodb.transform.TransformationInjector'
        ) as mock_injector:
            with mock.patch(
                'boto3.dynamodb.transform.DocumentModifiedShape.'
                'replace_documentation_for_matching_shape'
            ) as mock_modify_documentation_method:
                dynamodb_class(client=self.client)

        # It should have fired the following events upon instantiation.
        event_call_args = self.events.register.call_args_list
        assert event_call_args == [
            mock.call(
                'provide-client-params.dynamodb',
                copy_dynamodb_params,
                unique_id='dynamodb-create-params-copy',
            ),
            mock.call(
                'before-parameter-build.dynamodb',
                mock_injector.return_value.inject_condition_expressions,
                unique_id='dynamodb-condition-expression',
            ),
            mock.call(
                'before-parameter-build.dynamodb',
                mock_injector.return_value.inject_attribute_value_input,
                unique_id='dynamodb-attr-value-input',
            ),
            mock.call(
                'after-call.dynamodb',
                mock_injector.return_value.inject_attribute_value_output,
                unique_id='dynamodb-attr-value-output',
            ),
            mock.call(
                'docs.*.dynamodb.*.complete-section',
                mock_modify_documentation_method,
                unique_id='dynamodb-attr-value-docs',
            ),
            mock.call(
                'docs.*.dynamodb.*.complete-section',
                mock_modify_documentation_method,
                unique_id='dynamodb-key-expression-docs',
            ),
            mock.call(
                'docs.*.dynamodb.*.complete-section',
                mock_modify_documentation_method,
                unique_id='dynamodb-cond-expression-docs',
            ),
        ]


class TestRegisterHighLevelInterface(unittest.TestCase):
    def test_register(self):
        base_classes = [object]
        register_high_level_interface(base_classes)

        # Check that the base classes are as expected
        assert base_classes == [DynamoDBHighLevelResource, object]



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import botocore.session
import pytest

import boto3.session
from boto3 import utils
from tests import unittest


class TestUtils(unittest.TestCase):
    def test_runtime_error_raised_when_shadowing_client_method(self):
        botocore_session = botocore.session.get_session()
        session = boto3.session.Session(
            region_name='us-west-2', botocore_session=botocore_session
        )

        def shadows_put_object(class_attributes, **kwargs):
            utils.inject_attribute(class_attributes, 'put_object', 'invalid')

        botocore_session.register('creating-client-class', shadows_put_object)

        with pytest.raises(RuntimeError):
            # This should raise an exception because we're trying to
            # shadow the put_object client method in the
            # shadows_put_object handler above.
            session.client('s3')



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from boto3.docs.utils import get_resource_ignore_params
from boto3.resources.model import Parameter
from tests import unittest


class TestGetResourceIgnoreParams(unittest.TestCase):
    def test_target_is_single_resource(self):
        param = Parameter('InstanceId', 'response')
        ignore_params = get_resource_ignore_params([param])
        assert ignore_params == ['InstanceId']

    def test_target_is_multiple_resources(self):
        param = Parameter('InstanceIds[]', 'response')
        ignore_params = get_resource_ignore_params([param])
        assert ignore_params == ['InstanceIds']

    def test_target_is_element_of_multiple_resources(self):
        param = Parameter('InstanceIds[0]', 'response')
        ignore_params = get_resource_ignore_params([param])
        assert ignore_params == ['InstanceIds']

    def test_target_is_nested_param(self):
        param = Parameter('Filters[0].Name', 'response')
        ignore_params = get_resource_ignore_params([param])
        assert ignore_params == ['Filters']

        param = Parameter('Filters[0].Values[0]', 'response')
        ignore_params = get_resource_ignore_params([param])
        assert ignore_params == ['Filters']



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.docs.utils import DocumentedShape
from botocore.hooks import HierarchicalEmitter

from boto3.docs.method import document_model_driven_resource_method
from boto3.resources.model import ResponseResource
from tests.unit.docs import BaseDocsTest


class TestDocumentModelDrivenResourceMethod(BaseDocsTest):
    def setUp(self):
        super().setUp()
        self.event_emitter = HierarchicalEmitter()
        self.service_model = self.client.meta.service_model
        self.operation_model = self.service_model.operation_model(
            'SampleOperation'
        )
        self.service_resource_model = self.resource.meta.resource_model

    def test_default(self):
        document_model_driven_resource_method(
            self.doc_structure,
            'foo',
            self.operation_model,
            event_emitter=self.event_emitter,
            method_description='This describes the foo method.',
            example_prefix='response = myservice.foo',
            resource_action_model=self.service_resource_model.actions[0],
        )
        self.assert_contains_lines_in_order(
            [
                '.. py:method:: foo(**kwargs)',
                '  This describes the foo method.',
                '  **Request Syntax** ',
                '  ::',
                '    response = myservice.foo(',
                "        Foo='string',",
                "        Bar='string'",
                '    )',
                '  :type Foo: string',
                '  :param Foo: Documents Foo',
                '  :type Bar: string',
                '  :param Bar: Documents Bar',
                '  :rtype: dict',
                '  :returns: ',
                '    **Response Syntax** ',
                '    ::',
                '      {',
                "          'Foo': 'string',",
                "          'Bar': 'string'",
                '      }',
                '    **Response Structure** ',
                '    - *(dict) --* ',
                '      - **Foo** *(string) --* Documents Foo',
                '      - **Bar** *(string) --* Documents Bar',
            ]
        )

    def test_returns_resource(self):
        resource_action = self.service_resource_model.actions[0]
        # Override the return type of the action to be a resource
        # instead of the regular dictionary.
        resource_action.resource = ResponseResource(
            {
                'type': 'Sample',
                'identifiers': [
                    {
                        'target': 'Name',
                        'source': 'requestParameter',
                        'path': 'Foo',
                    }
                ],
            },
            self.resource_json_model,
        )
        document_model_driven_resource_method(
            self.doc_structure,
            'foo',
            self.operation_model,
            event_emitter=self.event_emitter,
            method_description='This describes the foo method.',
            example_prefix='sample = myservice.foo',
            resource_action_model=resource_action,
        )
        self.assert_contains_lines_in_order(
            [
                '.. py:method:: foo(**kwargs)',
                '  This describes the foo method.',
                '  **Request Syntax** ',
                '  ::',
                '    sample = myservice.foo(',
                "        Foo='string',",
                "        Bar='string'",
                '    )',
                '  :type Foo: string',
                '  :param Foo: Documents Foo',
                '  :type Bar: string',
                '  :param Bar: Documents Bar',
                '  :rtype: :py:class:`myservice.Sample`',
                '  :returns: Sample resource',
            ]
        )

    def test_returns_list_of_resource(self):
        resource_action = self.service_resource_model.actions[1]
        document_model_driven_resource_method(
            self.doc_structure,
            'foo',
            self.operation_model,
            event_emitter=self.event_emitter,
            method_description='This describes the foo method.',
            example_prefix='samples = myservice.foo',
            resource_action_model=resource_action,
        )
        self.assert_contains_lines_in_order(
            [
                '.. py:method:: foo(**kwargs)',
                '  This describes the foo method.',
                '  **Request Syntax** ',
                '  ::',
                '    samples = myservice.foo(',
                "        Foo='string',",
                "        Bar='string'",
                '    )',
                '  :type Foo: string',
                '  :param Foo: Documents Foo',
                '  :type Bar: string',
                '  :param Bar: Documents Bar',
                '  :rtype: list(:py:class:`myservice.Sample`)',
                '  :returns: A list of Sample resource',
            ]
        )

    def test_include_input(self):
        include_params = [
            DocumentedShape(
                name='Biz', type_name='string', documentation='biz docs'
            )
        ]
        document_model_driven_resource_method(
            self.doc_structure,
            'foo',
            self.operation_model,
            event_emitter=self.event_emitter,
            method_description='This describes the foo method.',
            example_prefix='response = myservice.foo',
            include_input=include_params,
            resource_action_model=self.service_resource_model.actions[0],
        )
        self.assert_contains_lines_in_order(
            [
                '.. py:method:: foo(**kwargs)',
                '  This describes the foo method.',
                '  **Request Syntax** ',
                '  ::',
                '    response = myservice.foo(',
                "        Foo='string',",
                "        Bar='string',",
                "        Biz='string'",
                '    )',
                '  :type Foo: string',
                '  :param Foo: Documents Foo',
                '  :type Bar: string',
                '  :param Bar: Documents Bar',
                '  :type Biz: string',
                '  :param Biz: biz docs',
                '  :rtype: dict',
                '  :returns: ',
                '    **Response Syntax** ',
                '    ::',
                '      {',
                "          'Foo': 'string',",
                "          'Bar': 'string'",
                '      }',
                '    **Response Structure** ',
                '    - *(dict) --* ',
                '      - **Foo** *(string) --* Documents Foo',
                '      - **Bar** *(string) --* Documents Bar',
            ]
        )

    def test_include_output(self):
        include_params = [
            DocumentedShape(
                name='Biz', type_name='string', documentation='biz docs'
            )
        ]
        document_model_driven_resource_method(
            self.doc_structure,
            'foo',
            self.operation_model,
            event_emitter=self.event_emitter,
            method_description='This describes the foo method.',
            example_prefix='response = myservice.foo',
            include_output=include_params,
            resource_action_model=self.service_resource_model.actions[0],
        )
        self.assert_contains_lines_in_order(
            [
                '.. py:method:: foo(**kwargs)',
                '  This describes the foo method.',
                '  **Request Syntax** ',
                '  ::',
                '    response = myservice.foo(',
                "        Foo='string',",
                "        Bar='string'",
                '    )',
                '  :type Foo: string',
                '  :param Foo: Documents Foo',
                '  :type Bar: string',
                '  :param Bar: Documents Bar',
                '  :rtype: dict',
                '  :returns: ',
                '    **Response Syntax** ',
                '    ::',
                '      {',
                "          'Foo': 'string',",
                "          'Bar': 'string',",
                "          'Biz': 'string'",
                '      }',
                '    **Response Structure** ',
                '    - *(dict) --* ',
                '      - **Foo** *(string) --* Documents Foo',
                '      - **Bar** *(string) --* Documents Bar',
                '      - **Biz** *(string) --* biz docs',
            ]
        )

    def test_exclude_input(self):
        document_model_driven_resource_method(
            self.doc_structure,
            'foo',
            self.operation_model,
            event_emitter=self.event_emitter,
            method_description='This describes the foo method.',
            example_prefix='response = myservice.foo',
            exclude_input=['Bar'],
            resource_action_model=self.service_resource_model.actions[0],
        )
        self.assert_contains_lines_in_order(
            [
                '.. py:method:: foo(**kwargs)',
                '  This describes the foo method.',
                '  **Request Syntax** ',
                '  ::',
                '    response = myservice.foo(',
                "        Foo='string',",
                '    )',
                '  :type Foo: string',
                '  :param Foo: Documents Foo',
                '  :rtype: dict',
                '  :returns: ',
                '    **Response Syntax** ',
                '    ::',
                '      {',
                "          'Foo': 'string',",
                "          'Bar': 'string'",
                '      }',
                '    **Response Structure** ',
                '    - *(dict) --* ',
                '      - **Foo** *(string) --* Documents Foo',
                '      - **Bar** *(string) --* Documents Bar',
            ]
        )
        self.assert_not_contains_lines(
            [':param Bar: string', 'Bar=\'string\'']
        )

    def test_exclude_output(self):
        document_model_driven_resource_method(
            self.doc_structure,
            'foo',
            self.operation_model,
            event_emitter=self.event_emitter,
            method_description='This describes the foo method.',
            example_prefix='response = myservice.foo',
            exclude_output=['Bar'],
            resource_action_model=self.service_resource_model.actions[0],
        )
        self.assert_contains_lines_in_order(
            [
                '.. py:method:: foo(**kwargs)',
                '  This describes the foo method.',
                '  **Request Syntax** ',
                '  ::',
                '    response = myservice.foo(',
                "        Foo='string',",
                "        Bar='string'",
                '    )',
                '  :type Foo: string',
                '  :param Foo: Documents Foo',
                '  :type Bar: string',
                '  :param Bar: Documents Bar',
                '  :rtype: dict',
                '  :returns: ',
                '    **Response Syntax** ',
                '    ::',
                '      {',
                "          'Foo': 'string'",
                '      }',
                '    **Response Structure** ',
                '    - *(dict) --* ',
                '      - **Foo** *(string) --* Documents Foo',
            ]
        )
        self.assert_not_contains_lines(
            [
                '\'Bar\': \'string\'',
                '- **Bar** *(string) --*',
            ]
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import types

import pytest

from boto3 import utils
from tests import mock, unittest


class FakeModule:
    @staticmethod
    def entry_point(**kwargs):
        return kwargs


class TestUtils(unittest.TestCase):
    def test_lazy_call(self):
        with mock.patch('boto3.utils.import_module') as importer:
            importer.return_value = FakeModule
            lazy_function = utils.lazy_call(
                'fakemodule.FakeModule.entry_point'
            )
            assert lazy_function(a=1, b=2) == {'a': 1, 'b': 2}

    def test_import_module(self):
        module = utils.import_module('boto3.s3.transfer')
        assert module.__name__ == 'boto3.s3.transfer'
        assert isinstance(module, types.ModuleType)

    def test_inject_attributes_with_no_shadowing(self):
        class_attributes = {}
        utils.inject_attribute(class_attributes, 'foo', 'bar')
        assert class_attributes['foo'] == 'bar'

    def test_shadowing_existing_var_raises_exception(self):
        class_attributes = {'foo': 'preexisting'}
        with pytest.raises(RuntimeError):
            utils.inject_attribute(class_attributes, 'foo', 'bar')


class TestLazyLoadedWaiterModel(unittest.TestCase):
    def test_get_waiter_model_is_lazy(self):
        session = mock.Mock()
        waiter_model = utils.LazyLoadedWaiterModel(
            session, 'myservice', '2014-01-01'
        )
        assert not session.get_waiter_model.called
        waiter_model.get_waiter('Foo')
        assert session.get_waiter_model.called
        session.get_waiter_model.return_value.get_waiter.assert_called_with(
            'Foo'
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from boto3.docs.client import Boto3ClientDocumenter
from tests.unit.docs import BaseDocsTest


class TestBoto3ClientDocumenter(BaseDocsTest):
    def setUp(self):
        super().setUp()
        self.client_documenter = Boto3ClientDocumenter(self.client)

    def test_document_client(self):
        self.client_documenter.document_client(self.doc_structure)
        self.assert_contains_lines_in_order(
            [
                '======',
                'Client',
                '======',
                '.. py:class:: MyService.Client',
                '  A low-level client representing AWS MyService',
                '  ::',
                '    import boto3',
                '    client = boto3.client(\'myservice\')',
                '  These are the available methods:',
                '  *   :py:meth:`~MyService.Client.can_paginate`',
                '  *   :py:meth:`~MyService.Client.get_paginator`',
                '  *   :py:meth:`~MyService.Client.get_waiter`',
                '  *   :py:meth:`~MyService.Client.sample_operation`',
                '  .. py:method:: can_paginate(operation_name)',
                '  .. py:method:: get_paginator(operation_name)',
                '  .. py:method:: get_waiter(waiter_name)',
                '  .. py:method:: sample_operation(**kwargs)',
                '    **Request Syntax**',
                '    ::',
                '      response = client.sample_operation(',
                '          Foo=\'string\'',
                '          Bar=\'string\'',
                '      )',
                '    :type Foo: string',
                '    :param Foo: Documents Foo',
                '    :type Bar: string',
                '    :param Bar: Documents Bar',
                '    :rtype: dict',
                '    :returns:',
                '      **Response Syntax**',
                '      ::',
                '        {',
                '            \'Foo\': \'string\'',
                '            \'Bar\': \'string\'',
                '        }',
                '      **Response Structure**',
                '      - *(dict) --*',
                '        - **Foo** *(string) --*',
                '        - **Bar** *(string) --*',
            ]
        )



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License'). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the 'license' file accompanying this file. This file is
# distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from boto3.ec2.deletetags import delete_tags
from tests import mock, unittest


class TestDeleteTags(unittest.TestCase):
    def setUp(self):
        self.client = mock.Mock()
        self.resource = mock.Mock()
        self.resource.meta.client = self.client
        self.instance_id = 'instance_id'
        self.resource.id = self.instance_id

    def test_delete_tags(self):
        tags = {
            'Tags': [
                {'Key': 'key1', 'Value': 'value1'},
                {'Key': 'key2', 'Value': 'value2'},
                {'Key': 'key3', 'Value': 'value3'},
            ]
        }

        delete_tags(self.resource, **tags)

        kwargs = tags
        kwargs['Resources'] = [self.instance_id]
        self.client.delete_tags.assert_called_with(**kwargs)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime
import hashlib
import io
import logging
import math
import os
import shutil
import string
import tempfile
import threading

from botocore.client import Config

import boto3.s3.transfer
import boto3.session
from tests import unique_id, unittest

LOG = logging.getLogger('boto3.tests.integration')


def assert_files_equal(first, second):
    if os.path.getsize(first) != os.path.getsize(second):
        raise AssertionError(f"Files are not equal: {first}, {second}")
    first_md5 = md5_checksum(first)
    second_md5 = md5_checksum(second)
    if first_md5 != second_md5:
        raise AssertionError(
            "Files are not equal: {}(md5={}) != {}(md5={})".format(
                first, first_md5, second, second_md5
            )
        )


def md5_checksum(filename):
    checksum = hashlib.md5()
    with open(filename, 'rb') as f:
        for chunk in iter(lambda: f.read(8192), b''):
            checksum.update(chunk)
    return checksum.hexdigest()


def random_bucket_name(prefix='boto3-transfer', num_chars=10):
    base = string.ascii_lowercase + string.digits
    random_bytes = bytearray(os.urandom(num_chars))
    return prefix + ''.join([base[b % len(base)] for b in random_bytes])


_SHARED_BUCKET = random_bucket_name()
_DEFAULT_REGION = 'us-west-2'


def setup_module():
    s3 = boto3.client('s3')
    waiter = s3.get_waiter('bucket_exists')
    params = {
        'Bucket': _SHARED_BUCKET,
        'CreateBucketConfiguration': {
            'LocationConstraint': _DEFAULT_REGION,
        },
    }
    try:
        s3.create_bucket(**params)
    except Exception as e:
        # A create_bucket can fail for a number of reasons.
        # We're going to defer to the waiter below to make the
        # final call as to whether or not the bucket exists.
        LOG.debug("create_bucket() raised an exception: %s", e, exc_info=True)
    waiter.wait(Bucket=_SHARED_BUCKET)


def clear_out_bucket(bucket, region, delete_bucket=False):
    s3 = boto3.client('s3', region_name=region)
    page = s3.get_paginator('list_objects')
    # Use pages paired with batch delete_objects().
    for page in page.paginate(Bucket=bucket):
        keys = [{'Key': obj['Key']} for obj in page.get('Contents', [])]
        if keys:
            s3.delete_objects(Bucket=bucket, Delete={'Objects': keys})
    if delete_bucket:
        try:
            s3.delete_bucket(Bucket=bucket)
        except Exception as e:
            # We can sometimes get exceptions when trying to
            # delete a bucket.  We'll let the waiter make
            # the final call as to whether the bucket was able
            # to be deleted.
            LOG.debug(
                "delete_bucket() raised an exception: %s", e, exc_info=True
            )
            waiter = s3.get_waiter('bucket_not_exists')
            waiter.wait(Bucket=bucket)


def teardown_module():
    clear_out_bucket(_SHARED_BUCKET, _DEFAULT_REGION, delete_bucket=True)


class FileCreator:
    def __init__(self):
        self.rootdir = tempfile.mkdtemp()

    def remove_all(self):
        shutil.rmtree(self.rootdir)

    def create_file(self, filename, contents, mode='w'):
        """Creates a file in a tmpdir

        ``filename`` should be a relative path, e.g. "foo/bar/baz.txt"
        It will be translated into a full path in a tmp dir.

        ``mode`` is the mode the file should be opened either as ``w`` or
        `wb``.

        Returns the full path to the file.

        """
        full_path = os.path.join(self.rootdir, filename)
        if not os.path.isdir(os.path.dirname(full_path)):
            os.makedirs(os.path.dirname(full_path))
        with open(full_path, mode) as f:
            f.write(contents)
        return full_path

    def create_file_with_size(self, filename, filesize):
        filename = self.create_file(filename, contents='')
        chunksize = 8192
        with open(filename, 'wb') as f:
            for i in range(int(math.ceil(filesize / float(chunksize)))):
                f.write(b'a' * chunksize)
        return filename

    def append_file(self, filename, contents):
        """Append contents to a file

        ``filename`` should be a relative path, e.g. "foo/bar/baz.txt"
        It will be translated into a full path in a tmp dir.

        Returns the full path to the file.
        """
        full_path = os.path.join(self.rootdir, filename)
        if not os.path.isdir(os.path.dirname(full_path)):
            os.makedirs(os.path.dirname(full_path))
        with open(full_path, 'a') as f:
            f.write(contents)
        return full_path

    def full_path(self, filename):
        """Translate relative path to full path in temp dir.

        f.full_path('foo/bar.txt') -> /tmp/asdfasd/foo/bar.txt
        """
        return os.path.join(self.rootdir, filename)


class TestS3Resource(unittest.TestCase):
    def setUp(self):
        self.region = _DEFAULT_REGION
        self.bucket_name = _SHARED_BUCKET
        clear_out_bucket(self.bucket_name, self.region)
        self.session = boto3.session.Session(region_name=self.region)
        self.s3 = self.session.resource('s3')
        self.bucket = self.s3.Bucket(self.bucket_name)

    def create_bucket_resource(self, bucket_name=None, region=None):
        if bucket_name is None:
            bucket_name = random_bucket_name()

        if region is None:
            region = self.region

        kwargs = {'Bucket': bucket_name}

        if region != 'us-east-1':
            kwargs['CreateBucketConfiguration'] = {
                'LocationConstraint': region
            }
        bucket = self.s3.create_bucket(**kwargs)
        self.addCleanup(bucket.delete)

        for _ in range(3):
            bucket.wait_until_exists()
        return bucket

    def test_s3(self):
        client = self.s3.meta.client

        # Create an object
        obj = self.bucket.Object('test.txt')
        obj.put(Body='hello, world')
        waiter = client.get_waiter('object_exists')
        waiter.wait(Bucket=self.bucket_name, Key='test.txt')
        self.addCleanup(obj.delete)

        # List objects and make sure ours is present
        self.assertIn('test.txt', [o.key for o in self.bucket.objects.all()])

        # Lazy-loaded attribute
        self.assertEqual(12, obj.content_length)

        # Load a similar attribute from the collection response
        self.assertEqual(12, list(self.bucket.objects.all())[0].size)

        # Perform a resource action with a low-level response
        self.assertEqual(b'hello, world', obj.get()['Body'].read())

    def test_s3_resource_waiter(self):
        # Create a bucket
        bucket_name = random_bucket_name()
        bucket = self.create_bucket_resource(bucket_name)
        # Wait till the bucket exists
        bucket.wait_until_exists()
        # Confirm the bucket exists by finding it in a list of all of our
        # buckets
        self.assertIn(bucket_name, [b.name for b in self.s3.buckets.all()])

        # Create an object
        obj = bucket.Object('test.txt')
        obj.put(Body='hello, world')
        self.addCleanup(obj.delete)

        # Wait till the bucket exists
        obj.wait_until_exists()

        # List objects and make sure ours is present
        self.assertIn('test.txt', [o.key for o in bucket.objects.all()])

    def test_can_create_object_directly(self):
        obj = self.s3.Object(self.bucket_name, 'test.txt')

        self.assertEqual(obj.bucket_name, self.bucket_name)
        self.assertEqual(obj.key, 'test.txt')

    def test_s3_multipart(self):
        # Create the multipart upload
        mpu = self.bucket.Object('mp-test.txt').initiate_multipart_upload()
        self.addCleanup(mpu.abort)

        # Create and upload a part
        part = mpu.Part(1)
        response = part.upload(Body='hello, world!')

        # Complete the upload, which requires info on all of the parts
        part_info = {'Parts': [{'PartNumber': 1, 'ETag': response['ETag']}]}

        mpu.complete(MultipartUpload=part_info)
        self.addCleanup(self.bucket.Object('mp-test.txt').delete)

        contents = self.bucket.Object('mp-test.txt').get()['Body'].read()
        self.assertEqual(contents, b'hello, world!')

    def test_s3_batch_delete(self):
        bucket = self.create_bucket_resource()
        bucket.Versioning().enable()

        # Create several versions of an object
        obj = self.bucket.Object('test.txt')
        for i in range(10):
            obj.put(Body="Version %s" % i)

            # Delete all the versions of the object
            bucket.object_versions.all().delete()

        versions = list(bucket.object_versions.all())
        self.assertEqual(len(versions), 0)


class TestS3Transfers(unittest.TestCase):
    """Tests for the high level boto3.s3.transfer module."""

    def setUp(self):
        self.region = _DEFAULT_REGION
        self.bucket_name = _SHARED_BUCKET
        clear_out_bucket(self.bucket_name, self.region)
        self.session = boto3.session.Session(region_name=self.region)
        self.client = self.session.client('s3', self.region)
        self.files = FileCreator()
        self.progress = 0

    def tearDown(self):
        self.files.remove_all()

    def delete_object(self, key):
        self.client.delete_object(Bucket=self.bucket_name, Key=key)

    def object_exists(self, key):
        waiter = self.client.get_waiter('object_exists')
        waiter.wait(Bucket=self.bucket_name, Key=key)
        return True

    def wait_until_object_exists(
        self, key_name, extra_params=None, min_successes=3
    ):
        waiter = self.client.get_waiter('object_exists')
        params = {'Bucket': self.bucket_name, 'Key': key_name}
        if extra_params is not None:
            params.update(extra_params)
        for _ in range(min_successes):
            waiter.wait(**params)

    def create_s3_transfer(self, config=None):
        return boto3.s3.transfer.S3Transfer(self.client, config=config)

    def assert_has_public_read_acl(self, response):
        grants = response['Grants']
        public_read = [
            g['Grantee'].get('URI', '')
            for g in grants
            if g['Permission'] == 'READ'
        ]
        self.assertIn('groups/global/AllUsers', public_read[0])

    def test_copy(self):
        self.client.put_object(
            Bucket=self.bucket_name, Key='foo', Body='beach'
        )
        self.addCleanup(self.delete_object, 'foo')

        self.client.copy(
            CopySource={'Bucket': self.bucket_name, 'Key': 'foo'},
            Bucket=self.bucket_name,
            Key='bar',
        )
        self.addCleanup(self.delete_object, 'bar')

        self.object_exists('bar')

    def test_upload_fileobj(self):
        fileobj = io.BytesIO(b'foo')
        self.client.upload_fileobj(
            Fileobj=fileobj, Bucket=self.bucket_name, Key='foo'
        )
        self.addCleanup(self.delete_object, 'foo')

        self.object_exists('foo')

    def test_upload_fileobj_progress(self):
        # This has to be an integration test because the fileobj will never
        # actually be read from when using the stubber and therefore the
        # progress callbacks will not be invoked.
        chunksize = 5 * (1024**2)
        config = boto3.s3.transfer.TransferConfig(
            multipart_chunksize=chunksize,
            multipart_threshold=chunksize,
            max_concurrency=1,
        )
        fileobj = io.BytesIO(b'0' * (chunksize * 3))

        def progress_callback(amount):
            self.progress += amount

        self.client.upload_fileobj(
            Fileobj=fileobj,
            Bucket=self.bucket_name,
            Key='foo',
            Config=config,
            Callback=progress_callback,
        )
        self.addCleanup(self.delete_object, 'foo')

        self.object_exists('foo')
        self.assertEqual(self.progress, chunksize * 3)

    def test_download_fileobj(self):
        fileobj = io.BytesIO()
        self.client.put_object(
            Bucket=self.bucket_name, Key='foo', Body=b'beach'
        )
        self.addCleanup(self.delete_object, 'foo')

        self.wait_until_object_exists('foo')
        self.client.download_fileobj(
            Bucket=self.bucket_name, Key='foo', Fileobj=fileobj
        )

        self.assertEqual(fileobj.getvalue(), b'beach')

    def test_upload_below_threshold(self):
        config = boto3.s3.transfer.TransferConfig(
            multipart_threshold=2 * 1024 * 1024
        )
        transfer = self.create_s3_transfer(config)
        filename = self.files.create_file_with_size(
            'foo.txt', filesize=1024 * 1024
        )
        transfer.upload_file(filename, self.bucket_name, 'foo.txt')
        self.addCleanup(self.delete_object, 'foo.txt')

        self.assertTrue(self.object_exists('foo.txt'))

    def test_upload_above_threshold(self):
        config = boto3.s3.transfer.TransferConfig(
            multipart_threshold=2 * 1024 * 1024
        )
        transfer = self.create_s3_transfer(config)
        filename = self.files.create_file_with_size(
            '20mb.txt', filesize=20 * 1024 * 1024
        )
        transfer.upload_file(filename, self.bucket_name, '20mb.txt')
        self.addCleanup(self.delete_object, '20mb.txt')
        self.assertTrue(self.object_exists('20mb.txt'))

    def test_upload_file_above_threshold_with_acl(self):
        config = boto3.s3.transfer.TransferConfig(
            multipart_threshold=5 * 1024 * 1024
        )
        transfer = self.create_s3_transfer(config)
        filename = self.files.create_file_with_size(
            '6mb.txt', filesize=6 * 1024 * 1024
        )
        extra_args = {'ACL': 'public-read'}
        transfer.upload_file(
            filename, self.bucket_name, '6mb.txt', extra_args=extra_args
        )
        self.addCleanup(self.delete_object, '6mb.txt')

        self.assertTrue(self.object_exists('6mb.txt'))
        response = self.client.get_object_acl(
            Bucket=self.bucket_name, Key='6mb.txt'
        )
        self.assert_has_public_read_acl(response)

    def test_upload_file_above_threshold_with_ssec(self):
        key_bytes = os.urandom(32)
        extra_args = {
            'SSECustomerKey': key_bytes,
            'SSECustomerAlgorithm': 'AES256',
        }
        config = boto3.s3.transfer.TransferConfig(
            multipart_threshold=5 * 1024 * 1024
        )
        transfer = self.create_s3_transfer(config)
        filename = self.files.create_file_with_size(
            '6mb.txt', filesize=6 * 1024 * 1024
        )
        transfer.upload_file(
            filename, self.bucket_name, '6mb.txt', extra_args=extra_args
        )
        self.addCleanup(self.delete_object, '6mb.txt')
        # A head object will fail if it has a customer key
        # associated with it and it's not provided in the HeadObject
        # request so we can use this to verify our functionality.
        response = self.client.head_object(
            Bucket=self.bucket_name, Key='6mb.txt', **extra_args
        )
        self.assertEqual(response['SSECustomerAlgorithm'], 'AES256')

    def test_progress_callback_on_upload(self):
        self.amount_seen = 0
        lock = threading.Lock()

        def progress_callback(amount):
            with lock:
                self.amount_seen += amount

        transfer = self.create_s3_transfer()
        filename = self.files.create_file_with_size(
            '20mb.txt', filesize=20 * 1024 * 1024
        )
        transfer.upload_file(
            filename, self.bucket_name, '20mb.txt', callback=progress_callback
        )
        self.addCleanup(self.delete_object, '20mb.txt')

        # The callback should have been called enough times such that
        # the total amount of bytes we've seen (via the "amount"
        # arg to the callback function) should be the size
        # of the file we uploaded.
        self.assertEqual(self.amount_seen, 20 * 1024 * 1024)

    def test_callback_called_once_with_sigv4(self):
        # Verify #98, where the callback was being invoked
        # twice when using signature version 4.
        self.amount_seen = 0
        lock = threading.Lock()

        def progress_callback(amount):
            with lock:
                self.amount_seen += amount

        client = self.session.client(
            's3', self.region, config=Config(signature_version='s3v4')
        )
        transfer = boto3.s3.transfer.S3Transfer(client)
        filename = self.files.create_file_with_size(
            '10mb.txt', filesize=10 * 1024 * 1024
        )
        transfer.upload_file(
            filename, self.bucket_name, '10mb.txt', callback=progress_callback
        )
        self.addCleanup(self.delete_object, '10mb.txt')

        self.assertEqual(self.amount_seen, 10 * 1024 * 1024)

    def test_can_send_extra_params_on_upload(self):
        transfer = self.create_s3_transfer()
        filename = self.files.create_file_with_size('foo.txt', filesize=1024)
        transfer.upload_file(
            filename,
            self.bucket_name,
            'foo.txt',
            extra_args={'ACL': 'public-read'},
        )
        self.addCleanup(self.delete_object, 'foo.txt')

        response = self.client.get_object_acl(
            Bucket=self.bucket_name, Key='foo.txt'
        )
        self.assert_has_public_read_acl(response)

    def test_can_configure_threshold(self):
        config = boto3.s3.transfer.TransferConfig(
            multipart_threshold=6 * 1024 * 1024
        )
        transfer = self.create_s3_transfer(config)
        filename = self.files.create_file_with_size(
            'foo.txt', filesize=8 * 1024 * 1024
        )
        transfer.upload_file(filename, self.bucket_name, 'foo.txt')
        self.addCleanup(self.delete_object, 'foo.txt')

        self.assertTrue(self.object_exists('foo.txt'))

    def test_can_send_extra_params_on_download(self):
        # We're picking the customer provided sse feature
        # of S3 to test the extra_args functionality of
        # S3.
        key_bytes = os.urandom(32)
        extra_args = {
            'SSECustomerKey': key_bytes,
            'SSECustomerAlgorithm': 'AES256',
        }
        self.client.put_object(
            Bucket=self.bucket_name,
            Key='foo.txt',
            Body=b'hello world',
            **extra_args,
        )
        self.addCleanup(self.delete_object, 'foo.txt')
        transfer = self.create_s3_transfer()

        download_path = os.path.join(self.files.rootdir, 'downloaded.txt')
        self.wait_until_object_exists('foo.txt', extra_params=extra_args)
        transfer.download_file(
            self.bucket_name, 'foo.txt', download_path, extra_args=extra_args
        )
        with open(download_path, 'rb') as f:
            self.assertEqual(f.read(), b'hello world')

    def test_progress_callback_on_download(self):
        self.amount_seen = 0
        lock = threading.Lock()

        def progress_callback(amount):
            with lock:
                self.amount_seen += amount

        transfer = self.create_s3_transfer()
        filename = self.files.create_file_with_size(
            '20mb.txt', filesize=20 * 1024 * 1024
        )
        with open(filename, 'rb') as f:
            self.client.put_object(
                Bucket=self.bucket_name, Key='20mb.txt', Body=f
            )
        self.addCleanup(self.delete_object, '20mb.txt')

        download_path = os.path.join(self.files.rootdir, 'downloaded.txt')
        transfer.download_file(
            self.bucket_name,
            '20mb.txt',
            download_path,
            callback=progress_callback,
        )

        self.assertEqual(self.amount_seen, 20 * 1024 * 1024)

    def test_download_below_threshold(self):
        transfer = self.create_s3_transfer()

        filename = self.files.create_file_with_size(
            'foo.txt', filesize=1024 * 1024
        )
        with open(filename, 'rb') as f:
            self.client.put_object(
                Bucket=self.bucket_name, Key='foo.txt', Body=f
            )
            self.addCleanup(self.delete_object, 'foo.txt')

        download_path = os.path.join(self.files.rootdir, 'downloaded.txt')
        self.wait_until_object_exists('foo.txt')
        transfer.download_file(self.bucket_name, 'foo.txt', download_path)
        assert_files_equal(filename, download_path)

    def test_download_above_threshold(self):
        transfer = self.create_s3_transfer()

        filename = self.files.create_file_with_size(
            'foo.txt', filesize=20 * 1024 * 1024
        )
        with open(filename, 'rb') as f:
            self.client.put_object(
                Bucket=self.bucket_name, Key='foo.txt', Body=f
            )
            self.addCleanup(self.delete_object, 'foo.txt')

        download_path = os.path.join(self.files.rootdir, 'downloaded.txt')
        self.wait_until_object_exists('foo.txt')
        transfer.download_file(self.bucket_name, 'foo.txt', download_path)
        assert_files_equal(filename, download_path)

    def test_download_file_with_directory_not_exist(self):
        transfer = self.create_s3_transfer()
        self.client.put_object(
            Bucket=self.bucket_name, Key='foo.txt', Body=b'foo'
        )
        self.addCleanup(self.delete_object, 'foo.txt')
        download_path = os.path.join(
            self.files.rootdir, 'a', 'b', 'c', 'downloaded.txt'
        )
        self.wait_until_object_exists('foo.txt')
        with self.assertRaises(IOError):
            transfer.download_file(self.bucket_name, 'foo.txt', download_path)

    def test_download_large_file_directory_not_exist(self):
        transfer = self.create_s3_transfer()

        filename = self.files.create_file_with_size(
            'foo.txt', filesize=20 * 1024 * 1024
        )
        with open(filename, 'rb') as f:
            self.client.put_object(
                Bucket=self.bucket_name, Key='foo.txt', Body=f
            )
            self.addCleanup(self.delete_object, 'foo.txt')
        download_path = os.path.join(
            self.files.rootdir, 'a', 'b', 'c', 'downloaded.txt'
        )
        self.wait_until_object_exists('foo.txt')
        with self.assertRaises(IOError):
            transfer.download_file(self.bucket_name, 'foo.txt', download_path)

    def test_transfer_methods_through_client(self):
        # This is really just a sanity check to ensure that the interface
        # from the clients work.  We're not exhaustively testing through
        # this client interface.
        filename = self.files.create_file_with_size(
            'foo.txt', filesize=1024 * 1024
        )
        self.client.upload_file(
            Filename=filename, Bucket=self.bucket_name, Key='foo.txt'
        )
        self.addCleanup(self.delete_object, 'foo.txt')

        download_path = os.path.join(self.files.rootdir, 'downloaded.txt')
        self.wait_until_object_exists('foo.txt')
        self.client.download_file(
            Bucket=self.bucket_name, Key='foo.txt', Filename=download_path
        )
        assert_files_equal(filename, download_path)

    def test_transfer_methods_do_not_use_threads(self):
        # This is just a smoke test to make sure that
        # setting use_threads to False has no issues transferring files as
        # the non-threaded implementation is ran under the same integration
        # and functional tests in s3transfer as the normal threaded
        # implementation
        #
        # The methods used are arbitrary other than one of the methods
        # use ``boto3.s3.transfer.S3Transfer`` and the other should be
        # using ``s3transfer.manager.TransferManager`` directly
        content = b'my content'
        filename = self.files.create_file('myfile', content.decode('utf-8'))
        key = 'foo'
        config = boto3.s3.transfer.TransferConfig(use_threads=False)

        self.client.upload_file(
            Bucket=self.bucket_name, Key=key, Filename=filename, Config=config
        )
        self.addCleanup(self.delete_object, key)
        self.assertTrue(self.object_exists(key))

        fileobj = io.BytesIO()
        self.client.download_fileobj(
            Bucket=self.bucket_name, Key='foo', Fileobj=fileobj, Config=config
        )
        self.assertEqual(fileobj.getvalue(), content)

    def test_transfer_methods_through_bucket(self):
        # This is just a sanity check to ensure that the bucket interface work.
        key = 'bucket.txt'
        bucket = self.session.resource('s3').Bucket(self.bucket_name)
        filename = self.files.create_file_with_size(key, 1024 * 1024)
        bucket.upload_file(Filename=filename, Key=key)
        self.addCleanup(self.delete_object, key)
        download_path = os.path.join(self.files.rootdir, unique_id('foo'))
        bucket.download_file(Key=key, Filename=download_path)
        assert_files_equal(filename, download_path)

    def test_transfer_methods_through_object(self):
        # This is just a sanity check to ensure that the object interface work.
        key = 'object.txt'
        obj = self.session.resource('s3').Object(self.bucket_name, key)
        filename = self.files.create_file_with_size(key, 1024 * 1024)
        obj.upload_file(Filename=filename)
        self.addCleanup(self.delete_object, key)
        download_path = os.path.join(self.files.rootdir, unique_id('foo'))
        obj.download_file(Filename=download_path)
        assert_files_equal(filename, download_path)


class TestCustomS3BucketLoad(unittest.TestCase):
    def setUp(self):
        self.region = _DEFAULT_REGION
        self.bucket_name = _SHARED_BUCKET
        clear_out_bucket(self.bucket_name, self.region)
        self.session = boto3.session.Session(region_name=self.region)
        self.s3 = self.session.resource('s3')

    def test_can_access_buckets_creation_date(self):
        bucket = self.s3.Bucket(self.bucket_name)
        self.assertIsInstance(bucket.creation_date, datetime.datetime)



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import botocore.session

import boto3.session
from tests import unittest


class TestUserAgentCustomizations(unittest.TestCase):
    def setUp(self):
        self.botocore_session = botocore.session.get_session()
        self.session = boto3.session.Session(
            region_name='us-west-2', botocore_session=self.botocore_session
        )
        self.actual_user_agent = None
        self.botocore_session.register(
            'request-created', self.record_user_agent
        )

    def record_user_agent(self, request, **kwargs):
        self.actual_user_agent = request.headers['User-Agent']

    def test_client_user_agent(self):
        client = self.session.client('s3')
        client.list_buckets()
        self.assertIn('Boto3', self.actual_user_agent)
        self.assertIn('Botocore', self.actual_user_agent)
        self.assertIn('Python', self.actual_user_agent)
        # We should *not* have any mention of resource
        # when using clients directly.
        self.assertNotIn('Resource', self.actual_user_agent)

    def test_resource_user_agent_has_customization(self):
        resource = self.session.resource('s3')
        list(resource.buckets.all())
        # We should have customized the user agent for
        # resource calls with "Resource".
        self.assertTrue(self.actual_user_agent.endswith(' Resource'))



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License'). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the 'license' file accompanying this file. This file is
# distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest
from botocore.model import DenormalizedStructureBuilder, ServiceModel

from boto3.exceptions import ResourceLoadException
from boto3.resources.base import ServiceResource
from boto3.resources.collection import CollectionManager
from boto3.resources.factory import ResourceFactory
from boto3.utils import ServiceContext
from tests import BaseTestCase, mock


class BaseTestResourceFactory(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.emitter = mock.Mock()
        self.factory = ResourceFactory(self.emitter)

    def load(
        self,
        resource_name,
        resource_json_definition=None,
        resource_json_definitions=None,
        service_model=None,
    ):
        if resource_json_definition is None:
            resource_json_definition = {}
        if resource_json_definitions is None:
            resource_json_definitions = {}
        service_context = ServiceContext(
            service_name='test',
            resource_json_definitions=resource_json_definitions,
            service_model=service_model,
            service_waiter_model=None,
        )

        return self.factory.load_from_definition(
            resource_name=resource_name,
            single_resource_json_definition=resource_json_definition,
            service_context=service_context,
        )


class TestResourceFactory(BaseTestResourceFactory):
    def test_get_service_returns_resource_class(self):
        TestResource = self.load('test')
        assert ServiceResource in TestResource.__bases__

    def test_get_resource_returns_resource_class(self):
        QueueResource = self.load('Queue')
        assert ServiceResource in QueueResource.__bases__

    def test_factory_sets_service_name(self):
        QueueResource = self.load('Queue')
        assert QueueResource.meta.service_name == 'test'

    def test_factory_sets_identifiers(self):
        model = {
            'identifiers': [
                {'name': 'QueueUrl'},
                {'name': 'ReceiptHandle'},
            ],
        }

        MessageResource = self.load('Message', model)

        assert 'queue_url' in MessageResource.meta.identifiers
        assert 'receipt_handle' in MessageResource.meta.identifiers

    def test_identifiers_in_repr(self):
        model = {
            'identifiers': [
                {'name': 'QueueUrl'},
                {'name': 'ReceiptHandle'},
            ],
        }
        defs = {'Message': model}

        resource = self.load('Message', model, defs)('url', 'handle')

        # Class name
        assert 'test.Message' in repr(resource)

        # Identifier names and values
        assert 'queue_url' in repr(resource)
        assert "'url'" in repr(resource)
        assert 'receipt_handle' in repr(resource)
        assert "'handle'" in repr(resource)

    def test_factory_creates_dangling_resources(self):
        model = {
            'has': {
                'Queue': {
                    'resource': {
                        'type': 'Queue',
                        'identifiers': [{'target': 'Url', 'source': 'input'}],
                    }
                },
                'Message': {
                    'resource': {
                        'type': 'Message',
                        'identifiers': [
                            {'target': 'QueueUrl', 'source': 'input'},
                            {'target': 'Handle', 'source': 'input'},
                        ],
                    }
                },
            }
        }
        defs = {'Queue': {}, 'Message': {}}

        TestResource = self.load('test', model, defs)

        assert hasattr(TestResource, 'Queue')
        assert hasattr(TestResource, 'Message')

    def test_factory_creates_properties(self):
        model = {
            'shape': 'TestShape',
            'load': {
                'request': {
                    'operation': 'DescribeTest',
                }
            },
        }
        shape = (
            DenormalizedStructureBuilder()
            .with_members(
                {
                    'ETag': {
                        'type': 'string',
                    },
                    'LastModified': {'type': 'string'},
                }
            )
            .build_model()
        )
        service_model = mock.Mock()
        service_model.shape_for.return_value = shape

        TestResource = self.load('test', model, service_model=service_model)

        assert hasattr(TestResource, 'e_tag')
        assert hasattr(TestResource, 'last_modified')

    def test_factory_renames_on_clobber_identifier(self):
        model = {'identifiers': [{'name': 'Meta'}]}

        # Each resource has a ``meta`` defined, so this identifier
        # must be renamed.
        cls = self.load('test', model)

        assert hasattr(cls, 'meta_identifier')

    def test_factory_fails_on_clobber_action(self):
        model = {
            'identifiers': [{'name': 'Test'}, {'name': 'TestAction'}],
            'actions': {'Test': {'request': {'operation': 'GetTest'}}},
        }

        # This fails because the resource has an identifier
        # that would be clobbered by the action name.
        with pytest.raises(ValueError) as cm:
            self.load('test', model)

            assert 'test' in str(cm.exception)
            assert 'action' in str(cm.exception)

    def test_can_instantiate_service_resource(self):
        TestResource = self.load('test')
        resource = TestResource()

        assert isinstance(resource, ServiceResource)

    def test_non_service_resource_missing_defs(self):
        # Only services should get dangling defs
        defs = {
            'Queue': {'identifiers': [{'name': 'Url'}]},
            'Message': {
                'identifiers': [
                    {'name': 'QueueUrl'},
                    {'name': 'ReceiptHandle'},
                ]
            },
        }

        model = defs['Queue']

        queue = self.load('Queue', model, defs)('url')

        assert not hasattr(queue, 'Queue')
        assert not hasattr(queue, 'Message')

    def test_subresource_requires_only_identifier(self):
        defs = {
            'Queue': {
                'identifiers': [{'name': 'Url'}],
                'has': {
                    'Message': {
                        'resource': {
                            'type': 'Message',
                            'identifiers': [
                                {
                                    'target': 'QueueUrl',
                                    'source': 'identifier',
                                    'name': 'Url',
                                },
                                {'target': 'ReceiptHandle', 'source': 'input'},
                            ],
                        }
                    }
                },
            },
            'Message': {
                'identifiers': [
                    {'name': 'QueueUrl'},
                    {'name': 'ReceiptHandle'},
                ]
            },
        }

        model = defs['Queue']

        queue = self.load('Queue', model, defs)('url')

        # Let's create a message and only give it a receipt handle
        # The required queue_url identifier should be set from the
        # queue itself.
        message = queue.Message('receipt')

        assert message.queue_url == 'url'
        assert message.receipt_handle == 'receipt'

    def test_resource_meta_unique(self):
        queue_cls = self.load('Queue')

        queue1 = queue_cls()
        queue2 = queue_cls()

        assert queue1.meta == queue2.meta

        queue1.meta.data = {'id': 'foo'}
        queue2.meta.data = {'id': 'bar'}

        assert queue_cls.meta != queue1.meta
        assert queue1.meta != queue2.meta
        assert queue1.meta != 'bad-value'

    def test_resource_meta_repr(self):
        queue_cls = self.load('Queue')
        queue = queue_cls()
        assert repr(queue.meta) == 'ResourceMeta(\'test\', identifiers=[])'

    @mock.patch('boto3.resources.factory.ServiceAction')
    def test_resource_calls_action(self, action_cls):
        model = {
            'actions': {
                'GetMessageStatus': {
                    'request': {'operation': 'DescribeMessageStatus'}
                }
            }
        }

        action = action_cls.return_value

        queue = self.load('Queue', model)()
        queue.get_message_status('arg1', arg2=2)

        action.assert_called_with(queue, 'arg1', arg2=2)

    @mock.patch('boto3.resources.factory.ServiceAction')
    def test_resource_action_clears_data(self, action_cls):
        model = {
            'load': {'request': {'operation': 'DescribeQueue'}},
            'actions': {
                'GetMessageStatus': {
                    'request': {'operation': 'DescribeMessageStatus'}
                }
            },
        }

        queue = self.load('Queue', model)()

        # Simulate loaded data
        queue.meta.data = {'some': 'data'}

        # Perform a call
        queue.get_message_status()

        # Cached data should be cleared
        assert queue.meta.data is None

    @mock.patch('boto3.resources.factory.ServiceAction')
    def test_resource_action_leaves_data(self, action_cls):
        # This model has NO load method. Cached data should
        # never be cleared since it cannot be reloaded!
        model = {
            'actions': {
                'GetMessageStatus': {
                    'request': {'operation': 'DescribeMessageStatus'}
                }
            }
        }

        queue = self.load('Queue', model)()

        # Simulate loaded data
        queue.meta.data = {'some': 'data'}

        # Perform a call
        queue.get_message_status()

        # Cached data should not be cleared
        assert queue.meta.data == {'some': 'data'}

    @mock.patch('boto3.resources.factory.ServiceAction')
    def test_resource_lazy_loads_properties(self, action_cls):
        model = {
            'shape': 'TestShape',
            'identifiers': [{'name': 'Url'}],
            'load': {
                'request': {
                    'operation': 'DescribeTest',
                }
            },
        }
        shape = (
            DenormalizedStructureBuilder()
            .with_members(
                {
                    'ETag': {'type': 'string', 'shape_name': 'ETag'},
                    'LastModified': {
                        'type': 'string',
                        'shape_name': 'LastModified',
                    },
                    'Url': {'type': 'string', 'shape_name': 'Url'},
                }
            )
            .build_model()
        )
        service_model = mock.Mock()
        service_model.shape_for.return_value = shape

        action = action_cls.return_value
        action.return_value = {'ETag': 'tag', 'LastModified': 'never'}

        resource = self.load('test', model, service_model=service_model)('url')

        # Accessing an identifier should not call load, even if it's in
        # the shape members.
        resource.url
        action.assert_not_called()

        # Accessing a property should call load
        assert resource.e_tag == 'tag'
        assert action.call_count == 1

        # Both params should have been loaded into the data bag
        assert 'ETag' in resource.meta.data
        assert 'LastModified' in resource.meta.data

        # Accessing another property should use cached value
        # instead of making a second call.
        assert resource.last_modified == 'never'
        assert action.call_count == 1

    @mock.patch('boto3.resources.factory.ServiceAction')
    def test_resource_lazy_properties_missing_load(self, action_cls):
        model = {
            'shape': 'TestShape',
            'identifiers': [{'name': 'Url'}]
            # Note the lack of a `load` method. These resources
            # are usually loaded via a call on a parent resource.
        }
        shape = (
            DenormalizedStructureBuilder()
            .with_members(
                {
                    'ETag': {
                        'type': 'string',
                    },
                    'LastModified': {'type': 'string'},
                    'Url': {'type': 'string'},
                }
            )
            .build_model()
        )
        service_model = mock.Mock()
        service_model.shape_for.return_value = shape

        action = action_cls.return_value
        action.return_value = {'ETag': 'tag', 'LastModified': 'never'}

        resource = self.load('test', model, service_model=service_model)('url')

        with pytest.raises(ResourceLoadException):
            resource.last_modified

    @mock.patch('boto3.resources.factory.ServiceAction')
    def test_resource_aliases_identifiers(self, action_cls):
        model = {
            'shape': 'TestShape',
            'identifiers': [{'name': 'id', 'memberName': 'foo_id'}],
        }
        shape = (
            DenormalizedStructureBuilder()
            .with_members(
                {
                    'foo_id': {
                        'type': 'string',
                    },
                    'bar': {'type': 'string'},
                }
            )
            .build_model()
        )
        service_model = mock.Mock()
        service_model.shape_for.return_value = shape

        shape_id = 'baz'
        resource = self.load('test', model, service_model=service_model)(
            shape_id
        )

        try:
            assert resource.id == shape_id
            assert resource.foo_id == shape_id
        except ResourceLoadException:
            self.fail("Load attempted on identifier alias.")

    def test_resource_loads_references(self):
        model = {
            'shape': 'InstanceShape',
            'identifiers': [{'name': 'GroupId'}],
            'has': {
                'Subnet': {
                    'resource': {
                        'type': 'Subnet',
                        'identifiers': [
                            {
                                'target': 'Id',
                                'source': 'data',
                                'path': 'SubnetId',
                            }
                        ],
                    }
                },
                'Vpcs': {
                    'resource': {
                        'type': 'Vpc',
                        'identifiers': [
                            {
                                'target': 'Id',
                                'source': 'data',
                                'path': 'Vpcs[].Id',
                            }
                        ],
                    }
                },
            },
        }
        defs = {
            'Subnet': {'identifiers': [{'name': 'Id'}]},
            'Vpc': {'identifiers': [{'name': 'Id'}]},
        }
        service_model = ServiceModel(
            {
                'shapes': {
                    'InstanceShape': {
                        'type': 'structure',
                        'members': {'SubnetId': {'shape': 'String'}},
                    },
                    'String': {'type': 'string'},
                }
            }
        )

        resource = self.load('Instance', model, defs, service_model)(
            'group-id'
        )

        # Load the resource with no data
        resource.meta.data = {}

        assert hasattr(resource, 'subnet')
        assert resource.subnet is None
        assert resource.vpcs is None

        # Load the resource with data to instantiate a reference
        resource.meta.data = {
            'SubnetId': 'abc123',
            'Vpcs': [{'Id': 'vpc1'}, {'Id': 'vpc2'}],
        }

        assert isinstance(resource.subnet, ServiceResource)
        assert resource.subnet.id == 'abc123'

        vpcs = resource.vpcs
        assert isinstance(vpcs, list)
        assert len(vpcs) == 2
        assert vpcs[0].id == 'vpc1'
        assert vpcs[1].id == 'vpc2'

    @mock.patch('boto3.resources.model.Collection')
    def test_resource_loads_collections(self, mock_model):
        model = {
            'hasMany': {
                'Queues': {
                    'request': {'operation': 'ListQueues'},
                    'resource': {'type': 'Queue'},
                }
            }
        }
        defs = {'Queue': {}}
        service_model = ServiceModel({})
        mock_model.return_value.name = 'queues'

        resource = self.load('test', model, defs, service_model)()

        # Resource must expose queues collection
        assert hasattr(resource, 'queues')
        assert isinstance(resource.queues, CollectionManager)

    def test_resource_loads_waiters(self):
        model = {
            "waiters": {
                "Exists": {
                    "waiterName": "BucketExists",
                    "params": [
                        {
                            "target": "Bucket",
                            "source": "identifier",
                            "name": "Name",
                        }
                    ],
                }
            }
        }

        defs = {'Bucket': {}}
        service_model = ServiceModel({})

        resource = self.load('test', model, defs, service_model)()

        assert hasattr(resource, 'wait_until_exists')

    @mock.patch('boto3.resources.factory.WaiterAction')
    def test_resource_waiter_calls_waiter_method(self, waiter_action_cls):
        model = {
            "waiters": {
                "Exists": {
                    "waiterName": "BucketExists",
                    "params": [
                        {
                            "target": "Bucket",
                            "source": "identifier",
                            "name": "Name",
                        }
                    ],
                }
            }
        }

        defs = {'Bucket': {}}
        service_model = ServiceModel({})

        waiter_action = waiter_action_cls.return_value
        resource = self.load('test', model, defs, service_model)()

        resource.wait_until_exists('arg1', arg2=2)
        waiter_action.assert_called_with(resource, 'arg1', arg2=2)


class TestResourceFactoryDanglingResource(BaseTestResourceFactory):
    def setUp(self):
        super().setUp()

        self.model = {
            'has': {
                'Queue': {
                    'resource': {
                        'type': 'Queue',
                        'identifiers': [{'target': 'Url', 'source': 'input'}],
                    }
                }
            }
        }

        self.defs = {'Queue': {'identifiers': [{'name': 'Url'}]}}

    def test_dangling_resources_create_resource_instance(self):
        resource = self.load('test', self.model, self.defs)()
        q = resource.Queue('test')

        assert isinstance(q, ServiceResource)

    def test_hash_resource_equal(self):
        resource = self.load('test', self.model, self.defs)()
        p = resource.Queue('test')
        q = resource.Queue('test')

        assert p == q
        assert hash(p) == hash(q)

    def test_hash_resource_not_equal(self):
        resource = self.load('test', self.model, self.defs)()
        p = resource.Queue('test1')
        q = resource.Queue('test2')

        assert p != q
        assert hash(p) != hash(q)

    def test_dangling_resource_create_with_kwarg(self):
        resource = self.load('test', self.model, self.defs)()
        q = resource.Queue(url='test')

        assert isinstance(q, ServiceResource)

    def test_dangling_resource_shares_client(self):
        resource = self.load('test', self.model, self.defs)()
        q = resource.Queue('test')

        assert resource.meta.client == q.meta.client

    def test_dangling_resource_requires_identifier(self):
        resource = self.load('test', self.model, self.defs)()

        with pytest.raises(ValueError):
            resource.Queue()

    def test_dangling_resource_raises_for_unknown_arg(self):
        resource = self.load('test', self.model, self.defs)()

        with pytest.raises(ValueError):
            resource.Queue(url='foo', bar='baz')

    def test_dangling_resource_identifier_is_immutable(self):
        resource = self.load('test', self.model, self.defs)()
        queue = resource.Queue('url')
        # We should not be able to change the identifier's value
        with pytest.raises(AttributeError):
            queue.url = 'foo'

    def test_dangling_resource_equality(self):
        resource = self.load('test', self.model, self.defs)()

        q1 = resource.Queue('url')
        q2 = resource.Queue('url')

        assert q1 == q2

    def test_dangling_resource_inequality(self):
        self.defs = {
            'Queue': {
                'identifiers': [{'name': 'Url'}],
                'has': {
                    'Message': {
                        'resource': {
                            'type': 'Message',
                            'identifiers': [
                                {
                                    'target': 'QueueUrl',
                                    'source': 'identifier',
                                    'name': 'Url',
                                },
                                {'target': 'Handle', 'source': 'input'},
                            ],
                        }
                    }
                },
            },
            'Message': {
                'identifiers': [{'name': 'QueueUrl'}, {'name': 'Handle'}]
            },
        }

        resource = self.load('test', self.model, self.defs)()

        q1 = resource.Queue('url')
        q2 = resource.Queue('different')
        m = q1.Message('handle')

        assert q1 != q2
        assert q1 != m

    def test_dangling_resource_loads_data(self):
        # Given a loadable resource instance that contains a reference
        # to another resource which has a resource data path, the
        # referenced resource should be loaded with all of the data
        # contained at that path. This allows loading references
        # which would otherwise not be loadable (missing load method)
        # and prevents extra load calls for others when we already
        # have the data available.
        self.defs = {
            'Instance': {
                'identifiers': [{'name': 'Id'}],
                'has': {
                    'NetworkInterface': {
                        'resource': {
                            'type': 'NetworkInterface',
                            'identifiers': [
                                {
                                    'target': 'Id',
                                    'source': 'data',
                                    'path': 'NetworkInterface.Id',
                                }
                            ],
                            'path': 'NetworkInterface',
                        }
                    }
                },
            },
            'NetworkInterface': {
                'identifiers': [{'name': 'Id'}],
                'shape': 'NetworkInterfaceShape',
            },
        }
        self.model = self.defs['Instance']
        shape = (
            DenormalizedStructureBuilder()
            .with_members(
                {
                    'Id': {
                        'type': 'string',
                    },
                    'PublicIp': {'type': 'string'},
                }
            )
            .build_model()
        )
        service_model = mock.Mock()
        service_model.shape_for.return_value = shape

        cls = self.load('Instance', self.model, self.defs, service_model)
        instance = cls('instance-id')

        # Set some data as if we had completed a load action.
        def set_meta_data():
            instance.meta.data = {
                'NetworkInterface': {
                    'Id': 'network-interface-id',
                    'PublicIp': '127.0.0.1',
                }
            }

        instance.load = mock.Mock(side_effect=set_meta_data)

        # Now, get the reference and make sure it has its data
        # set as expected.
        interface = instance.network_interface
        assert interface.meta.data is not None
        assert interface.public_ip == '127.0.0.1'


class TestServiceResourceSubresources(BaseTestResourceFactory):
    def setUp(self):
        super().setUp()

        self.model = {
            'has': {
                'QueueObject': {
                    'resource': {
                        'type': 'Queue',
                        'identifiers': [{'target': 'Url', 'source': 'input'}],
                    }
                },
                'PriorityQueue': {
                    'resource': {
                        'type': 'Queue',
                        'identifiers': [{'target': 'Url', 'source': 'input'}],
                    }
                },
            }
        }

        self.defs = {
            'Queue': {'identifiers': [{'name': 'Url'}]},
            'Message': {
                'identifiers': [
                    {'name': 'QueueUrl'},
                    {'name': 'ReceiptHandle'},
                ]
            },
        }

    def test_subresource_custom_name(self):
        resource = self.load('test', self.model, self.defs)()

        assert hasattr(resource, 'QueueObject')

    def test_contains_all_subresources(self):
        resource = self.load('test', self.model, self.defs)()

        assert 'QueueObject' in dir(resource)
        assert 'PriorityQueue' in dir(resource)
        assert 'Message' in dir(resource)

    def test_get_available_subresources(self):
        resource = self.load('test', self.model, self.defs)()
        assert hasattr(resource, 'get_available_subresources')
        subresources = sorted(resource.get_available_subresources())
        expected = sorted(['PriorityQueue', 'Message', 'QueueObject'])
        assert subresources == expected

    def test_subresource_missing_all_subresources(self):
        resource = self.load('test', self.model, self.defs)()
        message = resource.Message('url', 'handle')

        assert 'QueueObject' not in dir(message)
        assert 'PriorityQueue' not in dir(message)
        assert 'Queue' not in dir(message)
        assert 'Message' not in dir(message)

    def test_event_emitted_when_class_created(self):
        self.load('test', self.model, self.defs)
        assert self.emitter.emit.called
        call_args = self.emitter.emit.call_args
        # Verify the correct event name emitted.
        assert (
            call_args[0][0] == 'creating-resource-class.test.ServiceResource'
        )

        # Verify we send out the class attributes dict.
        actual_class_attrs = sorted(call_args[1]['class_attributes'])
        assert actual_class_attrs == [
            'Message',
            'PriorityQueue',
            'QueueObject',
            'get_available_subresources',
            'meta',
        ]

        base_classes = sorted(call_args[1]['base_classes'])
        assert base_classes == [ServiceResource]



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from boto3.docs.service import ServiceDocumenter
from boto3.session import Session
from tests.functional.docs import BaseDocsFunctionalTests


class TestDynamoDBCustomizations(BaseDocsFunctionalTests):
    def setUp(self):
        self.documenter = ServiceDocumenter(
            'dynamodb', session=Session(region_name='us-east-1')
        )
        self.generated_contents = self.documenter.document_service()
        self.generated_contents = self.generated_contents.decode('utf-8')

    def test_batch_writer_is_documented(self):
        self.assert_contains_lines_in_order(
            [
                '.. py:class:: DynamoDB.Table(name)',
                '  *   :py:meth:`batch_writer()`',
                '  .. py:method:: batch_writer(overwrite_by_pkeys=None)',
            ],
            self.generated_contents,
        )

    def test_document_interface_is_documented(self):
        contents = self.get_class_document_block(
            'DynamoDB.Table', self.generated_contents
        )

        # Take an arbitrary method that uses the customization.
        method_contents = self.get_method_document_block('put_item', contents)

        # Make sure the request syntax is as expected.
        request_syntax_contents = self.get_request_syntax_document_block(
            method_contents
        )
        self.assert_contains_lines_in_order(
            [
                'response = table.put_item(',
                'Item={',
                (
                    '\'string\': \'string\'|123|Binary(b\'bytes\')'
                    '|True|None|set([\'string\'])|set([123])|'
                    'set([Binary(b\'bytes\')])|[]|{}'
                ),
                '},',
                'Expected={',
                '\'string\': {',
                (
                    '\'Value\': \'string\'|123'
                    '|Binary(b\'bytes\')|True|None|set([\'string\'])'
                    '|set([123])|set([Binary(b\'bytes\')])|[]|{},'
                ),
                '\'AttributeValueList\': [',
                (
                    '\'string\'|123|Binary(b\'bytes\')'
                    '|True|None|set([\'string\'])|set([123])|'
                    'set([Binary(b\'bytes\')])|[]|{},'
                ),
            ],
            request_syntax_contents,
        )

        # Make sure the response syntax is as expected.
        response_syntax_contents = self.get_response_syntax_document_block(
            method_contents
        )
        self.assert_contains_lines_in_order(
            [
                '{',
                '\'Attributes\': {',
                (
                    '\'string\': \'string\'|123|'
                    'Binary(b\'bytes\')|True|None|set([\'string\'])|'
                    'set([123])|set([Binary(b\'bytes\')])|[]|{}'
                ),
                '},',
            ],
            response_syntax_contents,
        )

        # Make sure the request parameter is documented correctly.
        request_param_contents = self.get_request_parameter_document_block(
            'Item', method_contents
        )
        self.assert_contains_lines_in_order(
            [
                ':type Item: dict',
                ':param Item: **[REQUIRED]**',
                '- *(string) --*',
                (
                    '- *(valid DynamoDB type) --* - The value of the '
                    'attribute. The valid value types are listed in the '
                    ':ref:`DynamoDB Reference Guide<ref_valid_dynamodb_types>`.'
                ),
            ],
            request_param_contents,
        )

        # Make sure the response parameter is documented correctly.
        response_param_contents = self.get_response_parameter_document_block(
            'Attributes', method_contents
        )
        self.assert_contains_lines_in_order(
            [
                '- **Attributes** *(dict) --*',
                '- *(string) --*',
                (
                    '- *(valid DynamoDB type) --* - The value of '
                    'the attribute. The valid value types are listed in the '
                    ':ref:`DynamoDB Reference Guide<ref_valid_dynamodb_types>`.'
                ),
            ],
            response_param_contents,
        )

    def test_conditions_is_documented(self):
        contents = self.get_class_document_block(
            'DynamoDB.Table', self.generated_contents
        )

        # Take an arbitrary method that uses the customization.
        method_contents = self.get_method_document_block('query', contents)

        # Make sure the request syntax is as expected.
        request_syntax_contents = self.get_request_syntax_document_block(
            method_contents
        )
        self.assert_contains_lines_in_order(
            [
                'response = table.query(',
                ('FilterExpression=Attr(\'myattribute\').' 'eq(\'myvalue\'),'),
                ('KeyConditionExpression=Key(\'mykey\')' '.eq(\'myvalue\'),'),
            ],
            request_syntax_contents,
        )

        # Make sure the request parameter is documented correctly.
        self.assert_contains_lines_in_order(
            [
                (
                    ':type FilterExpression: condition from '
                    ':py:class:`boto3.dynamodb.conditions.Attr` method'
                ),
                (
                    ':param FilterExpression: The condition(s) an '
                    'attribute(s) must meet. Valid conditions are listed in '
                    'the :ref:`DynamoDB Reference Guide<ref_dynamodb_conditions>`.'
                ),
                (
                    ':type KeyConditionExpression: condition from '
                    ':py:class:`boto3.dynamodb.conditions.Key` method'
                ),
                (
                    ':param KeyConditionExpression: The condition(s) a '
                    'key(s) must meet. Valid conditions are listed in the '
                    ':ref:`DynamoDB Reference Guide<ref_dynamodb_conditions>`.'
                ),
            ],
            method_contents,
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License'). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the 'license' file accompanying this file. This file is
# distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest
from s3transfer.futures import NonThreadedExecutor
from s3transfer.manager import TransferManager

from boto3.exceptions import RetriesExceededError, S3UploadFailedError
from boto3.s3.transfer import (
    KB,
    MB,
    ClientError,
    OSUtils,
    ProgressCallbackInvoker,
    S3Transfer,
    S3TransferRetriesExceededError,
    TransferConfig,
    create_transfer_manager,
)
from tests import mock, unittest


class TestCreateTransferManager(unittest.TestCase):
    def test_create_transfer_manager(self):
        client = object()
        config = TransferConfig()
        osutil = OSUtils()
        with mock.patch('boto3.s3.transfer.TransferManager') as manager:
            create_transfer_manager(client, config, osutil)
            assert manager.call_args == mock.call(client, config, osutil, None)

    def test_create_transfer_manager_with_no_threads(self):
        client = object()
        config = TransferConfig()
        config.use_threads = False
        with mock.patch('boto3.s3.transfer.TransferManager') as manager:
            create_transfer_manager(client, config)
            assert manager.call_args == mock.call(
                client, config, None, NonThreadedExecutor
            )


class TestTransferConfig(unittest.TestCase):
    def assert_value_of_actual_and_alias(
        self, config, actual, alias, ref_value
    ):
        # Ensure that the name set in the underlying TransferConfig (i.e.
        # the actual) is the correct value.
        assert getattr(config, actual) == ref_value
        # Ensure that backcompat name (i.e. the alias) is the correct value.
        assert getattr(config, alias) == ref_value

    def test_alias_max_concurreny(self):
        ref_value = 10
        config = TransferConfig(max_concurrency=ref_value)
        self.assert_value_of_actual_and_alias(
            config, 'max_request_concurrency', 'max_concurrency', ref_value
        )

        # Set a new value using the alias
        new_value = 15
        config.max_concurrency = new_value
        # Make sure it sets the value for both the alias and the actual
        # value that will be used in the TransferManager
        self.assert_value_of_actual_and_alias(
            config, 'max_request_concurrency', 'max_concurrency', new_value
        )

    def test_alias_max_io_queue(self):
        ref_value = 10
        config = TransferConfig(max_io_queue=ref_value)
        self.assert_value_of_actual_and_alias(
            config, 'max_io_queue_size', 'max_io_queue', ref_value
        )

        # Set a new value using the alias
        new_value = 15
        config.max_io_queue = new_value
        # Make sure it sets the value for both the alias and the actual
        # value that will be used in the TransferManager
        self.assert_value_of_actual_and_alias(
            config, 'max_io_queue_size', 'max_io_queue', new_value
        )

    def test_transferconfig_parameters(self):
        config = TransferConfig(
            multipart_threshold=8 * MB,
            max_concurrency=10,
            multipart_chunksize=8 * MB,
            num_download_attempts=5,
            max_io_queue=100,
            io_chunksize=256 * KB,
            use_threads=True,
            max_bandwidth=1024 * KB,
        )
        assert config.multipart_threshold == 8 * MB
        assert config.multipart_chunksize == 8 * MB
        assert config.max_request_concurrency == 10
        assert config.num_download_attempts == 5
        assert config.max_io_queue_size == 100
        assert config.io_chunksize == 256 * KB
        assert config.use_threads is True
        assert config.max_bandwidth == 1024 * KB


class TestProgressCallbackInvoker(unittest.TestCase):
    def test_on_progress(self):
        callback = mock.Mock()
        subscriber = ProgressCallbackInvoker(callback)
        subscriber.on_progress(bytes_transferred=1)
        callback.assert_called_with(1)


class TestS3Transfer(unittest.TestCase):
    def setUp(self):
        self.client = mock.Mock()
        self.manager = mock.Mock(TransferManager(self.client))
        self.transfer = S3Transfer(manager=self.manager)
        self.callback = mock.Mock()

    def assert_callback_wrapped_in_subscriber(self, call_args):
        subscribers = call_args[0][4]
        # Make sure only one subscriber was passed in.
        assert len(subscribers) == 1
        subscriber = subscribers[0]
        # Make sure that the subscriber is of the correct type
        assert isinstance(subscriber, ProgressCallbackInvoker)
        # Make sure that the on_progress method() calls out to the wrapped
        # callback by actually invoking it.
        subscriber.on_progress(bytes_transferred=1)
        self.callback.assert_called_with(1)

    def test_upload_file(self):
        extra_args = {'ACL': 'public-read'}
        self.transfer.upload_file(
            'smallfile', 'bucket', 'key', extra_args=extra_args
        )
        self.manager.upload.assert_called_with(
            'smallfile', 'bucket', 'key', extra_args, None
        )

    def test_download_file(self):
        extra_args = {
            'SSECustomerKey': 'foo',
            'SSECustomerAlgorithm': 'AES256',
        }
        self.transfer.download_file(
            'bucket', 'key', '/tmp/smallfile', extra_args=extra_args
        )
        self.manager.download.assert_called_with(
            'bucket', 'key', '/tmp/smallfile', extra_args, None
        )

    def test_upload_wraps_callback(self):
        self.transfer.upload_file(
            'smallfile', 'bucket', 'key', callback=self.callback
        )
        self.assert_callback_wrapped_in_subscriber(
            self.manager.upload.call_args
        )

    def test_download_wraps_callback(self):
        self.transfer.download_file(
            'bucket', 'key', '/tmp/smallfile', callback=self.callback
        )
        self.assert_callback_wrapped_in_subscriber(
            self.manager.download.call_args
        )

    def test_propogation_of_retry_error(self):
        future = mock.Mock()
        future.result.side_effect = S3TransferRetriesExceededError(Exception())
        self.manager.download.return_value = future
        with pytest.raises(RetriesExceededError):
            self.transfer.download_file('bucket', 'key', '/tmp/smallfile')

    def test_propogation_s3_upload_failed_error(self):
        future = mock.Mock()
        future.result.side_effect = ClientError({'Error': {}}, 'op_name')
        self.manager.upload.return_value = future
        with pytest.raises(S3UploadFailedError):
            self.transfer.upload_file('smallfile', 'bucket', 'key')

    def test_can_create_with_just_client(self):
        transfer = S3Transfer(client=mock.Mock())
        assert isinstance(transfer, S3Transfer)

    def test_can_create_with_extra_configurations(self):
        transfer = S3Transfer(
            client=mock.Mock(), config=TransferConfig(), osutil=OSUtils()
        )
        assert isinstance(transfer, S3Transfer)

    def test_client_or_manager_is_required(self):
        with pytest.raises(ValueError):
            S3Transfer()

    def test_client_and_manager_are_mutually_exclusive(self):
        with pytest.raises(ValueError):
            S3Transfer(self.client, manager=self.manager)

    def test_config_and_manager_are_mutually_exclusive(self):
        with pytest.raises(ValueError):
            S3Transfer(config=mock.Mock(), manager=self.manager)

    def test_osutil_and_manager_are_mutually_exclusive(self):
        with pytest.raises(ValueError):
            S3Transfer(osutil=mock.Mock(), manager=self.manager)

    def test_upload_requires_string_filename(self):
        transfer = S3Transfer(client=mock.Mock())
        with pytest.raises(ValueError):
            transfer.upload_file(filename=object(), bucket='foo', key='bar')

    def test_download_requires_string_filename(self):
        transfer = S3Transfer(client=mock.Mock())
        with pytest.raises(ValueError):
            transfer.download_file(bucket='foo', key='bar', filename=object())

    def test_context_manager(self):
        manager = mock.Mock()
        manager.__exit__ = mock.Mock()
        with S3Transfer(manager=manager):
            pass
        # The underlying transfer manager should have had its __exit__
        # called as well.
        assert manager.__exit__.call_args == mock.call(None, None, None)

    def test_context_manager_with_errors(self):
        manager = mock.Mock()
        manager.__exit__ = mock.Mock()
        raised_exception = ValueError()
        with pytest.raises(type(raised_exception)):
            with S3Transfer(manager=manager):
                raise raised_exception
        # The underlying transfer manager should have had its __exit__
        # called as well and pass on the error as well.
        assert manager.__exit__.call_args == mock.call(
            type(raised_exception), raised_exception, mock.ANY
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from boto3.docs.collection import CollectionDocumenter
from tests.unit.docs import BaseDocsTest


class TestCollectionDocumenter(BaseDocsTest):
    def test_document_collections(self):
        collection_documenter = CollectionDocumenter(self.resource)
        collection_documenter.document_collections(self.doc_structure)
        self.assert_contains_lines_in_order(
            [
                '.. py:attribute:: samples',
                '  A collection of Sample resources.'
                'A Sample Collection will include all resources by default, '
                'and extreme caution should be taken when performing actions '
                'on all resources.',
                '  .. py:method:: all()',
                (
                    '    Creates an iterable of all Sample resources in the '
                    'collection.'
                ),
                '    **Request Syntax** ',
                '    ::',
                '      sample_iterator = myservice.samples.all()',
                '    :rtype: list(:py:class:`myservice.Sample`)',
                '    :returns: A list of Sample resources',
                '  .. py:method:: filter(**kwargs)',
                (
                    '    Creates an iterable of all Sample resources in '
                    'the collection filtered by kwargs passed to method. '
                    'A Sample collection will include all resources by default '
                    'if no filters are provided, and extreme caution should be '
                    'taken when performing actions on all resources'
                ),
                '    **Request Syntax** ',
                '    ::',
                '      sample_iterator = myservice.samples.filter(',
                "          Foo='string',",
                "          Bar='string'",
                '      )',
                '    :type Foo: string',
                '    :param Foo: Documents Foo',
                '    :type Bar: string',
                '    :param Bar: Documents Bar',
                '    :rtype: list(:py:class:`myservice.Sample`)',
                '    :returns: A list of Sample resources',
                '  .. py:method:: limit(**kwargs)',
                (
                    '    Creates an iterable up to a specified amount of '
                    'Sample resources in the collection.'
                ),
                '    **Request Syntax** ',
                '    ::',
                '      sample_iterator = myservice.samples.limit(',
                '          count=123',
                '      )',
                '    :type count: integer',
                (
                    '    :param count: The limit to the number of resources '
                    'in the iterable.'
                ),
                '    :rtype: list(:py:class:`myservice.Sample`)',
                '    :returns: A list of Sample resources',
                '  .. py:method:: operate(**kwargs)',
                '    **Request Syntax** ',
                '      response = myservice.samples.operate(',
                "          Foo='string',",
                "          Bar='string'",
                '      )',
                '    :type Foo: string',
                '    :param Foo: Documents Foo',
                '    :type Bar: string',
                '    :param Bar: Documents Bar',
                '    :rtype: dict',
                '    :returns: ',
                '      **Response Syntax** ',
                '      ::',
                '        {',
                "            'Foo': 'string',",
                "            'Bar': 'string'",
                '        }',
                '      **Response Structure** ',
                '      - *(dict) --* ',
                '        - **Foo** *(string) --* Documents Foo',
                '        - **Bar** *(string) --* Documents Bar',
                '  .. py:method:: page_size(**kwargs)',
                (
                    '    Creates an iterable of all Sample resources in the '
                    'collection, but limits the number of items returned by '
                    'each service call by the specified amount.'
                ),
                '    **Request Syntax** ',
                '    ::',
                '',
                '      sample_iterator = myservice.samples.page_size(',
                '          count=123',
                '      )',
                '    :type count: integer',
                (
                    '    :param count: The number of items returned by '
                    'each service call'
                ),
                '    :rtype: list(:py:class:`myservice.Sample`)',
                '    :returns: A list of Sample resources',
                '    ',
            ]
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from boto3.docs.waiter import WaiterResourceDocumenter
from tests.unit.docs import BaseDocsTest


class TestWaiterResourceDocumenter(BaseDocsTest):
    def test_document_resource_waiters(self):
        service_waiter_model = self.botocore_session.get_waiter_model(
            'myservice'
        )
        subresource = self.resource.Sample('mysample')
        waiter_documenter = WaiterResourceDocumenter(
            subresource, service_waiter_model
        )
        waiter_documenter.document_resource_waiters(self.doc_structure)
        self.assert_contains_lines_in_order(
            [
                '.. py:method:: wait_until_complete(**kwargs)',
                (
                    '  Waits until this Sample is complete. This method calls '
                    ':py:meth:`MyService.Waiter.sample_operation_complete.wait` '
                    'which polls. :py:meth:`MyService.Client.sample_operation` '
                    'every 15 seconds until a successful state is reached. An '
                    'error is returned after 40 failed checks.'
                ),
                '  **Request Syntax** ',
                '  ::',
                '    sample.wait_until_complete(',
                "        Bar='string'",
                '    )',
                '  :type Bar: string',
                '  :param Bar: Documents Bar',
                '  :returns: None',
            ]
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# https://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json

from botocore.stub import Stubber

import boto3
from tests import unittest


class TestTableResourceCustomizations(unittest.TestCase):

    maxDiff = None

    def setUp(self):
        self.resource = boto3.resource('dynamodb', 'us-east-1')
        self.table = self.resource.Table('mytable')

    def test_resource_has_batch_writer_added(self):
        assert hasattr(self.table, 'batch_writer')

    def test_operation_without_output(self):
        stubber = Stubber(self.table.meta.client)
        stubber.add_response('tag_resource', {})
        arn = 'arn:aws:dynamodb:us-west-2:123456789:table/mytable'

        with stubber:
            self.table.meta.client.tag_resource(
                ResourceArn=arn, Tags=[{'Key': 'project', 'Value': 'val'}]
            )

        stubber.assert_no_pending_responses()

    def test_batch_write_does_not_double_serialize(self):
        # If multiple items reference the same Python object, the
        # object does not get double-serialized.
        # https://github.com/boto/boto3/issues/3474

        used_twice = {'pkey': 'foo1', 'otherfield': {'foo': 1, 'bar': 2}}
        batch_writer = self.table.batch_writer()

        # The default Stubber compares the request payload to the
        # "expected_params" before automatic serialization happens. This custom
        # event handler uses the same technique as the Stubber to record the
        # serialized request body, but later in the request lifecycle.
        class LateStubber:
            def __init__(self, client):
                self.intercepted_request_body = None
                client.meta.events.register_first(
                    'before-call.*.*',
                    self.late_request_interceptor,
                )

            def late_request_interceptor(self, event_name, params, **kwargs):
                if self.intercepted_request_body is not None:
                    raise AssertionError(
                        'LateStubber was called more than once, but only one '
                        'request is expected'
                    )
                body_str = params.get('body', b'').decode('utf-8')
                try:
                    self.intercepted_request_body = json.loads(body_str)
                except Exception:
                    raise AssertionError(
                        'Expected JSON request body, but failed to JSON decode'
                    )

        late_stubber = LateStubber(self.table.meta.client)

        with Stubber(self.table.meta.client) as stubber:
            stubber.add_response(
                'batch_write_item',
                service_response={'UnprocessedItems': {}},
            )
            batch_writer.put_item(Item=used_twice)
            batch_writer.put_item(Item=used_twice)
            batch_writer._flush()

        expected_request_body = {
            'RequestItems': {
                'mytable': [
                    {
                        'PutRequest': {
                            'Item': {
                                'pkey': {'S': 'foo1'},
                                'otherfield': {
                                    'M': {'foo': {'N': '1'}, 'bar': {'N': '2'}}
                                },
                            }
                        }
                    },
                    {
                        'PutRequest': {
                            'Item': {
                                'pkey': {'S': 'foo1'},
                                'otherfield': {
                                    'M': {'foo': {'N': '1'}, 'bar': {'N': '2'}}
                                },
                            }
                        }
                    },
                ]
            }
        }

        assert late_stubber.intercepted_request_body == expected_request_body
